#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


def _instantiate(name):
    # Instantiate and return the specified class without cluttering up the
    # module with numerous imports.
    module_name, class_name = name.rsplit('.', 1)
    module = __import__(module_name, fromlist=[''])
    cls = module.__dict__[class_name]
    return cls()

cas_management = _instantiate('sasctl._services.cas_management.CASManagement')
concepts = _instantiate('sasctl._services.concepts.Concepts')
data_sources = _instantiate('sasctl._services.data_sources.DataSources')
files = _instantiate('sasctl._services.files.Files')
folders = _instantiate('sasctl._services.folders.Folders')
microanalytic_score = _instantiate(
    'sasctl._services.microanalytic_score.MicroAnalyticScore')
model_management = _instantiate(
    'sasctl._services.model_management.ModelManagement')
model_publish = _instantiate(
    'sasctl._services.model_publish.ModelPublish')
model_repository = _instantiate(
    'sasctl._services.model_repository.ModelRepository')
projects = _instantiate('sasctl._services.projects.Projects')
relationships = _instantiate('sasctl._services.relationships.Relationships')
reports = _instantiate('sasctl._services.reports.Reports')
report_images = _instantiate('sasctl._services.report_images.ReportImages')
sentiment_analysis = _instantiate(
    'sasctl._services.sentiment_analysis.SentimentAnalysis')
text_categorization = _instantiate(
    'sasctl._services.text_categorization.TextCategorization')
text_parsing = _instantiate('sasctl._services.text_parsing.TextParsing')





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service
from sasctl.utils.cli import sasctl_command


class Folders(Service):
    """The Folders API provides an organizational structure for SAS and
    external content. It can also be used for favorites folders or a history
    of objects accessed. The resources that are stored in folders (members)
    use a URI to point back to those resources.

    """
    _SERVICE_ROOT = '/folders'

    list_folders, get_folder, update_folder, \
    delete_folder = Service._crud_funcs('/folders', 'folder')

    @sasctl_command('folders', 'create')
    def create_folder(self, name, parent=None, description=None):
        """

        Parameters
        ----------
        name : str
            The name of the new folder
        parent : str or dict, optional
            The parent folder for this folder, if any.  Can be a folder name,
            id, or dict response from `get_folder`
        description : str, optional
            A description of the folder

        Returns
        -------

        """
        parent = self.get_folder(parent)

        body = {'name': name,
                'description': description,
                'folderType': 'folder',
                'parentFolderUri': parent.id if parent else None }

        return self.post('/folders',
                         json=body,
                         headers={'Content-Type': 'application/vnd.sas.content.folder+json'})


#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .service import Service
from ..core import uri_as_str

class SentimentAnalysis(Service):
    """The Sentiment Analysis API is used to perform sentiment analysis on
    natural language text documents according to a prebuilt or user-defined
    model.
    """
    _SERVICE_ROOT = '/sentiment'

    def analyze_sentiment(self,
                          documents,
                          caslib=None,
                          id_column=None,
                          text_column=None,
                          description=None,
                          model=None,
                          output_postfix=None,
                          language='en'):
        """Performs sentiment analysis on the input data.

        Creates a setiment analysis task that executes asynchronously.  There
        are two different interactions for sentiment analysis: analyzing
        documents in CAS tables and analyzing documents that are uploaded directly.

        Parameters
        ----------
        documents : str or dict or list_like:
            Documents to analyze.  May be either the URI to a CAS table where the
            documents are currently stored, or an iterable of strings containing
            the documents' text.
        caslib : str or dict, optional
            URI of a caslib in which the documents will be stored.  Required if
            `documents` is a list of strings.
        id_column : str, optional
            The column in `documents` that contains a unique id for each
            document.  Required if `documents` is a CAS table URI.
        text_column : str, optional
            The column in `documents` that contains the document text to
            analyze.  Required if `documents` is a CAS table URI.
        description : str, optional
            Description to add to the sentiment analysis job.
        model
        output_postfix : str, optional
            Text to be added to the end of all output table names.
        language : str, optional
            Two letter
            `ISO 639-1 <https://en.wikipedia.org/wiki/ISO_639>`_
            code indicating the source language.  Defaults to 'en'.

        Returns
        -------
        RestObj
            The submitted job

        See Also
        --------
        :meth:`cas_management.get_caslib <.CASManagement.get_caslib>`
        :meth:`cas_management.get_table <.CASManagement.get_table>`

        """
        if documents is None:
            raise TypeError('`documents` cannot be None.')

        if isinstance(documents, (dict, str)):
            data = {
                'inputUri': uri_as_str(documents),
                'documentIdVariable': id_column,
                'textVariable': text_column,
                'version': 1
            }
        else:
            data = {
                'caslibUri': uri_as_str(caslib),
                'documents': documents,
                'version': 1
            }

        data.update({
            'description': description,
            'language': language,
            'modelUri': uri_as_str(model),
            'outputTableNamePostfix': output_postfix
        })

        # Optional fields are not ignored if None. Explicitly remove before sending
        for k in list(data.keys()):
            if data[k] is None:
                del data[k]

        url = '/jobs'

        # Update URL if passing in raw documents.
        if 'documents' in data:
            url += '#data'
            headers = {
                'Content-Type':
                    'application/vnd.sas.text.sentiment.job.request.documents+json',
                'Accept': 'application/vnd.sas.text.sentiment.job+json'

            }
        else:
            headers = {
                'Content-Type':
                    'application/vnd.sas.text.sentiment.job.request+json',
                'Accept': 'application/vnd.sas.text.sentiment.job+json'
            }

        return self.post(url, json=data, headers=headers)


# Copyright (c) 2021, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .pickleModel import PickleModel
from .writeJSONFiles import JSONFiles
from .zipModel import ZipModel
from .writeScoreCode import ScoreCode
from .importModel import ImportModel



import os
import pickle
import xml.etree.ElementTree as etree
from io import StringIO

try:
    import xgboost
except ImportError:
    xgboost = None

try:
    import lightgbm
except ImportError:
    lightgbm = None

from sasctl.utils.decorators import experimental
from .connectors import LightgbmParser, PmmlParser, XgbParser


def _check_type(model):
    comp_types = [
        "xgboost.sklearn.XGBModel",
        "lightgbm.LGBMModel",
        "lightgbm.basic.Booster",
        "GBM.pmml file",
    ]

    if xgboost and isinstance(model, xgboost.sklearn.XGBModel):
        if model.booster not in ['gbtree', 'dart']:
            raise RuntimeError(
                "Model is xgboost. Unsupported booster type: %s."
                " Supported types are: %s" % (model.booster, ', '.join(comp_types))
            )

        parser = XgbParser(model.get_booster(), model.objective)
    elif lightgbm and isinstance(model, lightgbm.LGBMModel):
        parser = LightgbmParser(model.booster_)
    elif lightgbm and isinstance(model, lightgbm.basic.Booster):
        parser = LightgbmParser(model)
    elif etree and isinstance(model, etree.ElementTree):
        parser = PmmlParser(model.getroot())
    else:
        raise RuntimeError(
            "Unknown booster type: %s. Compatible types are: %s."
            " Check if corresponding library is installed."
            % (type(model).__name__, ', '.join(comp_types))
        )

    return parser


@experimental
def pyml2ds(in_file, out_var_name="P_TARGET"):
    """Translate a gradient boosting model and write SAS scoring code to file.

    Supported models are: xgboost, lightgbm and pmml gradient boosting.

    Parameters
    ----------
    in_file : str or bytes or file-like
        Pickled object to translate.  String is assumed to be a path to a picked
        file, file-like is assumed to be an open file handle to a pickle
        object, and bytes is assumed to be the raw pickled bytes.
    out_var_name : str (optional)
        Output variable name.

    Returns
    -------
    str
        A SAS Data Step program implementing the model.

    Examples
    --------
    Generate SAS code from an XGBoost model.

    >>> from xgboost.sklearn import XGBRegressor
    >>> xgb = XGBRegressor()
    >>> xgb.fit(X, y)
    >>> pkl = pickle.dumps(xgb)
    >>> sas_code = pyml2ds(pkl)

    """

    try:
        # In Python2 str could either be a path or the binary pickle data,
        # so check if its a valid filepath too.
        is_file_path = isinstance(in_file, str) and os.path.isfile(in_file)
    except TypeError:
        is_file_path = False

    # Path to a PMML or pickle file
    if is_file_path:
        # Parse PMML files
        if os.path.splitext(in_file)[-1] == '.pmml':
            model = etree.parse(in_file)
        else:
            # Read pickled files
            with open(in_file, 'rb') as f:
                model = pickle.load(f)

    elif isinstance(in_file, bytes):
        # Assume byte string is the actual pickled bytes
        model = pickle.loads(in_file)
    else:
        # Assume a file object containing the pickled object
        model = pickle.load(in_file)

    # Verify model is a valid type
    parser = _check_type(model)
    parser.out_var_name = out_var_name

    # Parser is currently written to expect a file input
    # Until refactored, use StringIO to collect the text in memory
    f = StringIO()
    parser.translate(f)

    # Return contents of "file"
    f.seek(0)
    return f.read()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .service import Service
from sasctl.core import uri_as_str


class TextParsing(Service):
    """The Text Parsing API parses natural language text documents.

    Parsing is a key operation in understanding your data. Parsing a document
    involves the following analyses:

     - Identifying terms used in the document
     - Recognizing parts of speech for each term
     - Identifying which terms are entities (person, country, and so on)
     - Resolving synonyms, misspellings, and so on

    The output tables that are generated during parsing can also be used in
    downstream analyses such as topic generation.

    """

    _SERVICE_ROOT = '/parsing'

    def parse_documents(
        self,
        documents,
        caslib=None,
        id_column=None,
        text_column=None,
        description=None,
        standard_entities=False,
        noun_groups=False,
        min_doc_count=10,
        concept_model=None,
        output_postfix=None,
        spell_check=False,
        override_list=None,
        stop_list=None,
        start_list=None,
        synonym_list=None,
        language='en',
    ):
        """Performs natural language parsing on the input data.

        Creates a text parsing job that executes asynchronously.  There are two
        different interactions for parsing: parsing documents in CAS tables and
        parsing documents that are uploaded directly.

        Parameters
        ----------
        documents : str or dict or list_like:
            Documents to parse.  May be either the URI to a CAS table where the
            documents are currently stored, or an iterable of strings containing
            the documents' text.
        caslib : str or dict, optional
            URI of a caslib in which the documents will be stored.  Required if
            `documents` is a list of strings.
        id_column : str, optional
            The column in `documents` that contains a unique id for each
            document.  Required if `documents` is a CAS table URI.
        text_column : str, optional
            The column in `documents` that contains the document text to parse.
            Required if `documents` is a CAS table URI.
        description : str, optional
            Description to add to the text parsing job.
        standard_entities : bool, optional
        noun_groups : bool, optional
        min_doc_count : int, optional
            Minimum number of documents in which a term must appear to be kept.
            Defaults to 10.
        output_postfix : str, optional
            Text to be added to the end of all output table names.
        spell_check : bool, optional
            Whether spell checking should be performed during parsing.
        concept_model : str or dict, optional
            URI of a table containing the concept LITI binaries to apply during
            parsing.
        override_list : str or dict, optional
            URI of a table containing overrides for the keep and drop terms.
        language : str, optional
            Two letter
            `ISO 639-1 <https://en.wikipedia.org/wiki/ISO_639>`_
            code indicating the source language.  Defaults to 'en'.

        Returns
        -------
        RestObj
            The submitted job

        See Also
        --------
        :meth:`cas_management.get_caslib <.CASManagement.get_caslib>`
        :meth:`cas_management.get_table <.CASManagement.get_table>`

        """

        if documents is None:
            raise TypeError('`documents` cannot be None.')

        if isinstance(documents, (dict, str)):
            data = {
                "inputUri": uri_as_str(documents),
                "documentIdVariable": id_column,
                "textVariable": text_column,
                "version": 1,
            }
        else:
            data = {
                'caslibUri': uri_as_str(caslib),
                'documents': documents,
                'version': 1,
            }

        data.update(
            {
                "description": description,
                "language": language,
                "includeStandardEntities": standard_entities,
                "includeNounGroups": noun_groups,
                "startListUri": uri_as_str(start_list),
                "stopListUri": uri_as_str(stop_list),
                "synonymListUri": uri_as_str(synonym_list),
                "minimumDocumentCount": min_doc_count,
                "conceptModelUri": uri_as_str(concept_model),
                "outputTableNamePostfix": output_postfix,
                "enableSpellChecking": spell_check,
                "overrideListUri": uri_as_str(override_list),
            }
        )

        # Optional fields are not ignored if None so explicitly remove before
        # sending.
        for k in list(data.keys()):
            if data[k] is None:
                del data[k]

        url = '/jobs'

        # Update URL if passing in raw documents.
        if 'documents' in data:
            url += '#data'
            headers = {
                'Content-Type': 'application/vnd.sas.text.parsing.job.request.documents+json',
                'Accept': 'application/vnd.sas.text.parsing.job+json',
            }
        else:
            headers = {
                'Content-Type': 'application/vnd.sas.text.parsing.job.request+json',
                'Accept': 'application/vnd.sas.text.parsing.job+json',
            }

        return self.post(url, json=data, headers=headers)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service


class CASManagement(Service):
    """The CAS Management service provides the ability to manage and perform
    actions on common resources as they relate to Cloud Analytic Services (CAS).
    """

    _SERVICE_ROOT = '/casManagement'

    list_servers, get_server, _, _ = Service._crud_funcs('/servers', 'server')

    def list_caslibs(self, server, filter=None):
        """List caslibs available on a server.

        Parameters
        ----------
        server : str or dict
            Name, ID, or dictionary representation of the server.
        filter : str, optional
            A `formatted <https://developer.sas.com/reference/filtering>`_
            filter string.

        Returns
        -------
        list
            A collection of :class:`.RestObj` instances.

        """
        return self._get_rel(server, 'caslibs', func=self.get_server,
                             filter=filter) or []

    def get_caslib(self, name, server=None):
        """Get a caslib by name.

        Parameters
        ----------
        name : str
            Name of the caslib
        server : str, optional
            Name of the CAS server.  Defaults to `cas-shared-default`.

        Returns
        -------
        RestObj

        """
        server = server or 'cas-shared-default'
        caslibs = self.list_caslibs(server,
                                    filter='eq($primary,name, "%s")' % name)

        if caslibs:
            return caslibs[0]

    def list_tables(self, caslib, server=None, filter=None):
        """List tables available in a caslib.

        Parameters
        ----------
        caslib : str or dict
            Name, ID, or dictionary representation of the caslib.
        server : str, optional
            Server where the `caslib` is registered.
        filter : str, optional
            Filter string in the `https://developer.sas.com/reference/filtering
            /` format.

        Returns
        -------
        list
            A collection of :class:`.RestObj` instances.

        """
        return self._get_rel(caslib, 'tables', self.get_caslib, filter,
                             server) or []

    def get_table(self, name, caslib, server=None):
        """Get a table by name.

        Parameters
        ----------
        name : str
            Name of the table.
        caslib : str or dict
            Name, ID, or dictionary representation of the caslib.
        server : str, optional
            Server where the `caslib` is registered.

        Returns
        -------
        RestObj

        """
        tables = self.list_tables(caslib,
                                  server=server,
                                  filter='eq($primary,name, "%s")' % name)

        if tables:
            return tables[0]



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service


class Projects(Service):
    _SERVICE_ROOT = '/projects'

    list_projects, get_project, update_project, \
        delete_project = Service._crud_funcs('/projects', 'project')

    def create_project(self, name, description=None, image=None):
        """

        Parameters
        ----------
        name : str
        description : str
        image : str
            URI of an image to use as the project avatar

        Returns
        -------
        RestObj

        """
        body = {'name': name,
                'description': description,
                'imageUri': image
                }

        return self.post('/projects', json=body,
                         headers={'Content-Type': 'application/vnd.sas.project+json'})


# Copyright (c) 2020, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from uuid import UUID
from warnings import warn

from ..core import platform_version
from .._services.model_repository import ModelRepository as mr
from .writeScoreCode import ScoreCode as sc
from .zipModel import ZipModel as zm


def project_exists(response, project):
    '''Checks if project exists on SAS Viya. If the project does not exist, then a new
    project is created or an error is raised.

    Parameters
    ----------
    response : dict
        JSON response of the get_project() call to model repository service.
    project : string or dict
        The name or id of the model project, or a dictionary representation of the project.

    Returns
    -------
    response : dict
        JSON response of the get_project() call to model repository service.

    Raises
    ------
    SystemError
        Alerts user that API calls cannot continue until a valid project is provided.
    '''
    if response is None:
        try:
            warn('No project with the name or UUID {} was found.'.format(project))
            UUID(project)
            raise SystemError(
                'The provided UUID does not match any projects found in SAS Model Manager. '
                + 'Please enter a valid UUID or a new name for a project to be created.'
            )
        except ValueError:
            repo = mr.default_repository().get('id')
            response = mr.create_project(project, repo)
            print('A new project named {} was created.'.format(response.name))
            return response
    else:
        return response


def model_exists(project, name, force):
    '''Checks if model already exists and either raises an error or deletes the redundant model.

    Parameters
    ----------
    project : string or dict
        The name or id of the model project, or a dictionary representation of the project.
    name : str or dict
        The name of the model.
    force : bool, optional
        Sets whether to overwrite models with the same name upon upload.

    Raises
    ------
    ValueError
        Model repository API cannot overwrite an already existing model with the upload model call.
        Alerts user of the force argument to allow multi-call API overwriting.
    '''
    project = mr.get_project(project)
    projectId = project['id']
    projectModels = mr.get('/projects/{}/models'.format(projectId))

    for model in projectModels:
        # Throws a TypeError if only one model is in the project
        try:
            if model['name'] == name:
                if force:
                    mr.delete_model(model.id)
                else:
                    raise ValueError(
                        'A model with the same model name exists in project {}. Include the force=True argument to overwrite models with the same name.'.format(
                            project.name
                        )
                    )
        except TypeError:
            if projectModels['name'] == name:
                if force:
                    mr.delete_model(projectModels.id)
                else:
                    raise ValueError(
                        'A model with the same model name exists in project {}. Include the force=True argument to overwrite models with the same name.'.format(
                            project.name
                        )
                    )


class ImportModel:
    @classmethod
    def pzmmImportModel(
        cls,
        zPath,
        modelPrefix,
        project,
        inputDF,
        targetDF,
        predictmethod,
        metrics=['EM_EVENTPROBABILITY', 'EM_CLASSIFICATION'],
        modelFileName=None,
        pyPath=None,
        threshPrediction=None,
        otherVariable=False,
        isH2OModel=False,
        force=False,
        binaryString=None,
    ):
        '''Import model to SAS Model Manager using pzmm submodule.

        Using pzmm, generate Python score code and import the model files into
        SAS Model Manager. This function automatically checks the version of SAS
        Viya being used through the sasctl Session object and creates the appropriate
        score code and API calls required for the model and its associated content to
        be registered in SAS Model Manager.

        The following are generated by this function:
        * '*Score.py'
            The Python score code file for the model.
        * '*.zip'
            The zip archive of the relevant model files. In Viya 3.5 the Python score
            code is not present in this initial zip file.


        Parameters
        ----------
        zPath : string or Path
            Directory location of the files to be zipped and imported as a model.
        modelPrefix : string
            The variable for the model name that is used when naming model files.
            (For example: hmeqClassTree + [Score.py || .pickle]).
        project : str or dict
            The name or id of the model project, or a dictionary
            representation of the project.
        inputDF : DataFrame
            The `DataFrame` object contains the training data, and includes only the predictor
            columns. The writeScoreCode function currently supports int(64), float(64),
            and string data types for scoring.
        targetDF : DataFrame
            The `DataFrame` object contains the training data for the target variable.
        predictMethod : string
            User-defined prediction method for score testing. This should be
            in a form such that the model and data input can be added using
            the format() command.
            For example: '{}.predict_proba({})'.
        metrics : string list, optional
            The scoring metrics for the model. The default is a set of two
            metrics: EM_EVENTPROBABILITY and EM_CLASSIFICATION.
        modelFileName : string, optional
            Name of the model file that contains the model. By default None and assigned as
            modelPrefix + '.pickle'.
        pyPath : string, optional
            The local path of the score code file. By default None and assigned as the zPath.
        threshPrediction : float, optional
            The prediction threshold for probability metrics. For classification,
            below this threshold is a 0 and above is a 1.
        otherVariable : boolean, optional
            The option for having a categorical other value for catching missing
            values or values not found in the training data set. The default setting
            is False.
        isH2OModel : boolean, optional
            Sets whether the model is an H2O.ai Python model. By default False.
        force : boolean, optional
            Sets whether to overwrite models with the same name upon upload. By default False.
        binaryString : string, optional
            Binary string representation of the model object. By default None.
        '''
        # Initialize no score code or binary H2O model flags
        noScoreCode = False
        binaryModel = False

        if pyPath is None:
            pyPath = Path(zPath)
        else:
            pyPath = Path(pyPath)

        # Function to check for MOJO or binary model files in H2O models
        def getFiles(extensions):
            allFiles = []
            for ext in extensions:
                allFiles.extend(pyPath.glob(ext))
            return allFiles

        # If the model file name is not provided, set a default value depending on H2O and binary model status
        if modelFileName is None:
            if isH2OModel:
                binaryOrMOJO = getFiles(['*.mojo', '*.pickle'])
                if len(binaryOrMOJO) == 0:
                    print(
                        'WARNING: An H2O model file was not found at {}. Score code will not be automatically generated.'.format(
                            str(pyPath)
                        )
                    )
                    noScoreCode = True
                elif len(binaryOrMOJO) == 1:
                    if str(binaryOrMOJO[0]).endswith('.pickle'):
                        binaryModel = True
                        modelFileName = modelPrefix + '.pickle'
                    else:
                        modelFileName = modelPrefix + '.mojo'
                else:
                    print(
                        'WARNING: Both a MOJO and binary model file are present at {}. Score code will not be automatically generated.'.format(
                            str(pyPath)
                        )
                    )
                    noScoreCode = True
            else:
                modelFileName = modelPrefix + '.pickle'

        # Check the SAS Viya version number being used
        isViya35 = platform_version() == '3.5'
        # For SAS Viya 4, the score code can be written beforehand and imported with all of the model files
        if not isViya35:
            if noScoreCode:
                print('No score code was generated.')
            else:
                sc.writeScoreCode(
                    inputDF,
                    targetDF,
                    modelPrefix,
                    predictmethod,
                    modelFileName,
                    metrics=metrics,
                    pyPath=pyPath,
                    threshPrediction=threshPrediction,
                    otherVariable=otherVariable,
                    isH2OModel=isH2OModel,
                    isBinaryModel=binaryModel,
                    binaryString=binaryString,
                )
                print(
                    'Model score code was written successfully to {}.'.format(
                        Path(pyPath) / (modelPrefix + 'Score.py')
                    )
                )
            zipIOFile = zm.zipFiles(Path(zPath), modelPrefix)
            print('All model files were zipped to {}.'.format(Path(zPath)))

            # Check if project name provided exists and raise an error or create a new project
            projectResponse = mr.get_project(project)
            project = project_exists(projectResponse, project)

            # Check if model with same name already exists in project.
            model_exists(project, modelPrefix, force)

            response = mr.import_model_from_zip(modelPrefix, project, zipIOFile)
            try:
                print(
                    'Model was successfully imported into SAS Model Manager as {} with UUID: {}.'.format(
                        response.name, response.id
                    )
                )
            except AttributeError:
                print('Model failed to import to SAS Model Manager.')
        # For SAS Viya 3.5, the score code is written after upload in order to know the model UUID
        else:
            zipIOFile = zm.zipFiles(Path(zPath), modelPrefix)
            print('All model files were zipped to {}.'.format(Path(zPath)))

            # Check if project name provided exists and raise an error or create a new project
            projectResponse = mr.get_project(project)
            project = project_exists(projectResponse, project)

            # Check if model with same name already exists in project.
            model_exists(project, modelPrefix, force)

            response = mr.import_model_from_zip(modelPrefix, project, zipIOFile, force)
            try:
                print(
                    'Model was successfully imported into SAS Model Manager as {} with UUID: {}.'.format(
                        response.name, response.id
                    )
                )
            except AttributeError:
                print('Model failed to import to SAS Model Manager.')
            if noScoreCode:
                print('No score code was generated.')
            else:
                sc.writeScoreCode(
                    inputDF,
                    targetDF,
                    modelPrefix,
                    predictmethod,
                    modelFileName,
                    metrics=metrics,
                    pyPath=pyPath,
                    threshPrediction=threshPrediction,
                    otherVariable=otherVariable,
                    model=response.id,
                    isH2OModel=isH2OModel,
                    isBinaryModel=binaryModel,
                    binaryString=binaryString,
                )
                print(
                    'Model score code was written successfully to {} and uploaded to SAS Model Manager'.format(
                        Path(pyPath) / (modelPrefix + 'Score.py')
                    )
                )



# Copyright (c) 2020, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# %%
from pathlib import Path
import sys

import swat
import getpass
import json
import pandas as pd
from sklearn import metrics
import numpy as np
from scipy.stats import kendalltau, gamma

# %%
class JSONFiles():
        
    def writeVarJSON(self, inputDF, isInput=True, jPath=Path.cwd()):
        '''
        Writes a variable descriptor JSON file for input or output variables,
        based on an input dataframe containing predictor and prediction columns.
        
        The following files are generated by this function:
        * 'inputVar.json' or 'outputVar.json'
            Output JSON file located at jPath.
        
        Parameters
        ---------------
        inputDF : Dataframe
            Input dataframe containing the training data set in a 
            pandas.Dataframe format. Columns are used to define predictor and
            prediction variables (ambiguously named "predict").
        isInput : boolean
            Boolean to check if generating the input or output variable JSON.
        jPath : string, optional
            File location for the output JSON file. Default is the current
            working directory.
        '''
        
        try:
            predictNames = inputDF.columns.values.tolist()
            isSeries = False
        except AttributeError:
            predictNames = [inputDF.name]
            isSeries = True
        outputJSON = pd.DataFrame()
        
        # loop through all predict variables to determine their name, length,
        # type, and level; append each to outputJSON
        for name in predictNames:
            if isSeries:
                predict = inputDF
            else:
                predict = inputDF[name]
            firstRow = predict.loc[predict.first_valid_index()]
            dType = predict.dtypes.name
            isStr = type(firstRow) is str
                
            if isStr:
                outputLevel = 'nominal'
                outputType = 'string'
                outputLength = predict.str.len().max()
            else:
                if dType == 'category':
                    outputLevel = 'nominal'
                else:
                    outputLevel = 'interval'
                outputType = 'decimal'
                outputLength = 8
                
            outputRow = pd.Series([name,
                                   outputLength,
                                   outputType,
                                   outputLevel],
                                  index=['name',
                                         'length',
                                         'type',
                                         'level']
                                  )
            outputJSON = outputJSON.append([outputRow], ignore_index=True)
        
        if isInput:
            fileName = 'inputVar.json'
        else:
            fileName = 'outputVar.json'
            
        with open(Path(jPath) / fileName, 'w') as jFile:
            dfDump = pd.DataFrame.to_dict(outputJSON.transpose()).values()
            json.dump(list(dfDump),
                      jFile,
                      indent=4,
                      skipkeys=True)
        print('{} was successfully written and saved to {}'.format(fileName, Path(jPath) / fileName))
            
    def writeModelPropertiesJSON(self, modelName, modelDesc, targetVariable,
                                 modelType, modelPredictors, targetEvent,
                                 numTargetCategories, eventProbVar=None,
                                 jPath=Path.cwd(), modeler=None):
        '''
        Writes a model properties JSON file. The JSON file format is required by the
        Model Repository API service and only eventProbVar can be 'None'.
        
        Parameters
        ---------------
        modelName : string
            User-defined model name.
        modelDesc : string
            User-defined model description.
        targetVariable : string
            Target variable to be predicted by the model.
        modelType : string
            User-defined model type.
        modelPredictors : string list
            List of predictor variables for the model.
        targetEvent : string
            Model target event (for example, 1 for a binary event).
        numTargetCategories : int
            Number of possible target categories (for example, 2 for a binary event).
        eventProbVar : string, optional
            Model prediction metric for scoring. Default is None.
        jPath : string, optional
            Location for the output JSON file. The default is the current
            working directory.
        modeler : string, optional
            The modeler name to be displayed in the model properties. The
            default value is None.
            
        Yields
        ---------------
        'ModelProperties.json'
            Output JSON file located at jPath.            
        '''
        
        # Check if model description provided is smaller than then 1024 character limit
        if len(modelDesc) > 1024:
            modelDesc = modelDesc[:1024]
            print('WARNING: The provided model description was truncated to 1024 characters.')
            
        if numTargetCategories > 2:
            targetLevel = 'NOMINAL'
        else:
            targetLevel = 'BINARY'
            
        if eventProbVar is None:
            try:
                eventProbVar = 'P_' + targetVariable + targetEvent
            except TypeError:
                eventProbVar = None
        # Replace <myUserID> with the user ID of the modeler that created the model.
        if modeler is None:
            try:
                modeler = getpass.getuser()
            except OSError:
                modeler = '<myUserID>'
        
        pythonVersion = sys.version.split(' ', 1)[0]
        
        propIndex = ['name', 'description', 'function',
                     'scoreCodeType', 'trainTable', 'trainCodeType',
                     'algorithm', 'targetVariable', 'targetEvent',
                     'targetLevel', 'eventProbVar', 'modeler',
                     'tool', 'toolVersion']
        
        modelProperties = [modelName, modelDesc, 'classification',
                           'python', ' ', 'Python',
                           modelType, targetVariable, targetEvent,
                           targetLevel, eventProbVar, modeler,
                           'Python 3', pythonVersion]
        
        outputJSON = pd.Series(modelProperties, index=propIndex)
        
        with open(Path(jPath) / 'ModelProperties.json', 'w') as jFile:
            dfDump = pd.Series.to_dict(outputJSON.transpose())
            json.dump(dfDump,
                      jFile,
                      indent=4,
                      skipkeys=True)
        print('{} was successfully written and saved to {}'.format('ModelProperties.json', Path(jPath) / 'ModelProperties.json'))
            
    def writeFileMetadataJSON(self, modelPrefix, jPath=Path.cwd(), isH2OModel=False):
        '''
        Writes a file metadata JSON file pointing to all relevant files.
        
        Parameters
        ---------------
        modelPrefix : string
            The variable for the model name that is used when naming model files.
            (for example, hmeqClassTree + [Score.py || .pickle]).
        jPath : string, optional
            Location for the output JSON file. The default value is the current
            working directory.   
        isH2OModel : boolean, optional
            Sets whether the model metadata is associated with an H2O.ai model.
            If set as True, the MOJO model file will be set as a score resource.
            The default value is False.
            
        Yields
        ---------------
        'fileMetadata.json'
            Output JSON file located at jPath.            
        '''
        
        if not isH2OModel:
            fileMetadata = pd.DataFrame([['inputVariables', 'inputVar.json'],
                                         ['outputVariables', 'outputVar.json'],
                                         ['score', modelPrefix + 'Score.py'],
                                         ['scoreResource',
                                          modelPrefix + '.pickle']],
                                        columns = ['role', 'name']
                                        )
        else:
            fileMetadata = pd.DataFrame([['inputVariables', 'inputVar.json'],
                                         ['outputVariables', 'outputVar.json'],
                                         ['score', modelPrefix + 'Score.py'],
                                         ['scoreResource',
                                          modelPrefix + '.mojo']],
                                        columns = ['role', 'name']
                                        )
        
        with open(Path(jPath) / 'fileMetadata.json', 'w') as jFile:
            dfDump = pd.DataFrame.to_dict(fileMetadata.transpose()).values()
            json.dump(list(dfDump),
                      jFile,
                      indent=4,
                      skipkeys=True)
        print('{} was successfully written and saved to {}'.format('fileMetaData.json', Path(jPath) / 'fileMetaData.json'))
            
    def writeBaseFitStat(self, csvPath=None, jPath=Path.cwd(),
                         userInput=False, tupleList=None):
        '''
        Writes a JSON file to display fit statistics for the model in SAS Open Model Manager.
        There are three modes to add fit parameters to the JSON file:
            
            1. Call the function with additional tuple arguments containing
            the name of the parameter, its value, and the partition that it 
            belongs to.
            
            2. Provide line by line user input prompted by the function.
            
            3. Import values from a CSV file. Format should contain the above
            tuple in each row.
            
        The following are the base statistical parameters SAS Model Manager
        and SAS Open Model Manager currently supports:
            * RASE = Root Average Squared Error
            * NObs = Sum of Frequencies
            * GINI = Gini Coefficient
            * GAMMA = Gamma
            * MCE = Misclassification Rate
            * ASE = Average Squared Error
            * MCLL = Multi-Class Log Loss
            * KS = KS (Youden)
            * KSPostCutoff = ROC Separation
            * DIV = Divisor for ASE
            * TAU = Tau
            * KSCut = KS Cutoff
            * C = Area Under ROC
        
        Parameters
        ---------------
        csvPath : string, optional
            Location for an input CSV file that contains parameter statistics.
            The default value is None.
        jPath : string, optional
            Location for the output JSON file. The default value is the current
            working directory.
        userInput : boolean, optional
            If true, prompt the user for more parameters. The default value is false.
        tupleList : list of tuples, optional
            Input parameter tuples in the form of (parameterName, 
            parameterLabel, parameterValue, dataRole). For example,
            a sample parameter call would be ('NObs', 'Sum of Frequencies',
            3488, 'TRAIN'). Variable dataRole is typically either TRAIN,
            TEST, or VALIDATE or 1, 2, 3 respectively. The default value is None.
        
        Yields
        ---------------
        'dmcas_fitstat.json'
            Output JSON file located at jPath.
        '''
        
        validParams = ['_RASE_', '_NObs_', '_GINI_', '_GAMMA_', '_MCE_',
                       '_ASE_', '_MCLL_', '_KS_', '_KSPostCutoff_', '_DIV_',
                       '_TAU_', '_KSCut_', '_C_']
        
        nullJSONPath = Path(__file__).resolve().parent / 'null_dmcas_fitstat.json'
        nullJSONDict = self.readJSONFile(nullJSONPath)
        
        dataMap = [{}, {}, {}]
        for i in range(3):
            dataMap[i] = nullJSONDict['data'][i]
        
        if tupleList is not None:
            for paramTuple in tupleList:
                # ignore incorrectly formatted input arguments
                if type(paramTuple) == tuple and len(paramTuple) == 3:
                    paramName = self.formatParameter(paramTuple[0])
                    if paramName not in validParams:
                        continue
                    if type(paramTuple[2]) == str:
                        dataRole = self.convertDataRole(paramTuple[2]) 
                    else:
                        dataRole = paramTuple[2]
                    dataMap[dataRole-1]['dataMap'][paramName] = paramTuple[1]
        
        if userInput:        
            while True:
                paramName = input('Parameter name: ')
                paramName = self.formatParameter(paramName)
                if paramName not in validParams:
                    print('Not a valid parameter. Please see documentation.')
                    if input('More parameters? (Y/N)') == 'N':
                        break
                    continue
                paramValue = input('Parameter value: ')
                dataRole = input('Data role: ')
                
                if type(dataRole) is str:
                    dataRole = self.convertDataRole(dataRole)
                dataMap[dataRole-1]['dataMap'][paramName] = paramValue
                
                if input('More parameters? (Y/N)') == 'N':
                    break
        
        if csvPath is not None:
            csvData = pd.read_csv(csvPath)
            for i, row in enumerate(csvData.values):
                paramName, paramValue, dataRole = row
                paramName = self.formatParameter(paramName)
                if paramName not in validParams:
                    continue
                if type(dataRole) is str:
                    dataRole = self.convertDataRole(dataRole)
                dataMap[dataRole-1]['dataMap'][paramName] = paramValue
                
        outJSON = nullJSONDict
        for i in range(3):
            outJSON['data'][i] = dataMap[i]
                
        with open(Path(jPath) / 'dmcas_fitstat.json', 'w') as jFile:
            json.dump(outJSON,
                      jFile,
                      indent=4,
                      skipkeys=True)
        print('{} was successfully written and saved to {}'.format('dmcas_fitstat.json', Path(jPath) / 'dmcas_fitstat.json'))
            
    def calculateFitStat(self, validateData=None, trainData=None, 
                         testData=None, jPath=Path.cwd()):
        '''
        Calculates fit statistics from user data and predictions and then writes to
        a JSON file for importing into the common model repository. Note that if
        no data set is provided (validate, train, or test), this function raises
        an error and does not create a JSON file.
        
        Data sets can be provided in the following forms:
        * pandas dataframe; the actual and predicted values are their own columns
        * numpy array; the actual and predicted values are their own columns or rows 
        and ordered such that the actual values come first and the predicted second
        * list; the actual and predicted values are their own indexed entry
        
        Datasets can be provided in the following forms:
        * pandas dataframe; the actual and predicted values are their own columns
        * numpy array; the actual and predicted values are their own columns or rows 
        and ordered such that the actual values come first and the predicted second
        * list; the actual and predicted values are their own indexed entry
        
        Datasets can be provided in the following forms:
        * pandas dataframe; the actual and predicted values are their own columns
        * numpy array; the actual and predicted values are their own columns or rows 
        and ordered such that the actual values come first and the predicted second
        * list; the actual and predicted values are their own indexed entry
        
        Parameters
        ---------------
        validateData : pandas dataframe, numpy array, or list, optional
            Dataframe, array, or list of the validation data set, including both
            the actual and predicted values. The default value is None.
        trainData : pandas dataframe, numpy array, or list, optional
            Dataframe, array, or list of the train data set, including both
            the actual and predicted values. The default value is None.
        testData : pandas dataframe, numpy array, or list, optional
            Dataframe, array, or list of the test data set, including both
            the actual and predicted values. The default value is None.
        jPath : string, optional
            Location for the output JSON file. The default value is the current
            working directory.
        
        Yields
        ---------------
        'dmcas_fitstat.json'
            Output JSON file located at jPath.
        '''
        
        nullJSONPath = Path(__file__).resolve().parent / 'null_dmcas_fitstat.json'
        nullJSONDict = self.readJSONFile(nullJSONPath)
        
        dataSets = [[[None], [None]],
                    [[None], [None]],
                    [[None], [None]]]
        
        dataPartitionExists = []
        for i, data in enumerate([validateData, trainData, testData]):
            if data is not None:
                dataPartitionExists.append(i)
                if type(data) is np.ndarray:
                    dataSets[i] = data.tolist()
                elif type(data) is pd.core.frame.DataFrame:
                    dataSets[i] = data.transpose().values.tolist()
                elif type(data) is list:
                    dataSets[i] = data
                    
        if len(dataPartitionExists) == 0:
            try:
                raise ValueError
            except ValueError:
                print('No data was provided. Please provide the actual ' +
                      'and predicted values for at least one of the ' + 
                      'partitions (VALIDATE, TRAIN, or TEST).')
                raise
                            
        for j in dataPartitionExists:
            fitStats = nullJSONDict['data'][j]['dataMap']

            fitStats['_PartInd_'] = j
            
            # If the data provided is Predicted | Actual instead of Actual | Predicted, catch the error and flip the columns
            try:
                fpr, tpr, _ = metrics.roc_curve(dataSets[j][0], dataSets[j][1])
            except ValueError:
                tempSet = dataSets[j]
                dataSets[j][0] = tempSet[1]
                dataSets[j][1] = tempSet[0]
                fpr, tpr, _ = metrics.roc_curve(dataSets[j][0], dataSets[j][1])
        
            RASE = np.sqrt(metrics.mean_squared_error(dataSets[j][0], dataSets[j][1]))
            fitStats['_RASE_'] = RASE
        
            NObs = len(dataSets[j][0])
            fitStats['_NObs_'] = NObs
        
            auc = metrics.roc_auc_score(dataSets[j][0], dataSets[j][1])
            GINI = (2 * auc) - 1
            fitStats['_GINI_'] = GINI
        
            _, _, scale = gamma.fit(dataSets[j][1])
            fitStats['_GAMMA_'] = 1/scale
        
            intPredict = [round(x) for x in dataSets[j][1]]
            MCE = 1 - metrics.accuracy_score(dataSets[j][0], intPredict)
            fitStats['_MCE_'] = MCE
        
            ASE = metrics.mean_squared_error(dataSets[j][0], dataSets[j][1])
            fitStats['_ASE_'] = ASE
        
            MCLL = metrics.log_loss(dataSets[j][0], dataSets[j][1])
            fitStats['_MCLL_'] = MCLL
        
            KS = max(np.abs(fpr-tpr))
            fitStats['_KS_'] = KS
        
            KSPostCutoff = None
            fitStats['_KSPostCutoff_'] = KSPostCutoff
        
            DIV = len(dataSets[j][0])
            fitStats['_DIV_'] = DIV
        
            TAU, _ = kendalltau(dataSets[j][0], dataSets[j][1])
            fitStats['_TAU_'] = TAU
        
            KSCut = None
            fitStats['_KSCut_'] = KSCut
        
            C = metrics.auc(fpr, tpr)
            fitStats['_C_'] = C
        
            nullJSONDict['data'][j]['dataMap'] = fitStats

        with open(Path(jPath) / 'dmcas_fitstat.json', 'w') as jFile:
            json.dump(nullJSONDict, jFile, indent=4)            
        print('{} was successfully written and saved to {}'.format('dmcas_fitstat.json', Path(jPath) / 'dmcas_fitstat.json'))
    
    def generateROCLiftStat(self, targetName, targetValue, swatConn, 
                            validateData=None, trainData=None, testData=None, 
                            jPath=Path.cwd()):
        '''
        Calculates the ROC and Lift curves from user data and model predictions and then 
        writes it to JSON files for importing in to the common model repository.
        ROC and Lift calculations are completed by CAS through a SWAT call. Note that if
        no data set is provided (validate, train, or test), this function raises
        an error and does not create any JSON files.
        
        Parameters
        ---------------
        targetName: str
            Target variable name to be predicted.
        targetValue: int or float
            Value of target variable that indicates an event.
        swatConn: SWAT connection to CAS
            Connection object to CAS service in SAS Model Manager or SAS® Open Model Manager through SWAT authentication.
        validateData : pandas dataframe, numpy array, or list, optional
            Dataframe, array, or list of the validation data set, including both
            the actual values and the calculated probabilities. The default value is None.
        trainData : pandas dataframe, numpy array, or list, optional
            Dataframe, array, or list of the train data set, including both
            the actual values and the calculated probabilities. The default value is None.
        testData : pandas dataframe, numpy array, or list, optional
            Dataframe, array, or list of the test data set, including both
            the actual values and the calculated probabilities. The default value is None.
        jPath : string, optional
            Location for the output JSON file. The default value is the current
            working directory.
        
        Yields
        ---------------
        'dmcas_roc.json'
            Output JSON file located at jPath.
        'dmcas_lift.json'
            Output JSON file located at jPath.
        '''
        
        nullJSONROCPath = Path(__file__).resolve().parent / 'null_dmcas_roc.json'
        nullJSONROCDict = self.readJSONFile(nullJSONROCPath)
        
        nullJSONLiftPath = Path(__file__).resolve().parent / 'null_dmcas_lift.json'
        nullJSONLiftDict = self.readJSONFile(nullJSONLiftPath)
                
        dataSets = [pd.DataFrame(), pd.DataFrame(), pd.DataFrame()]
        columns = ['actual', 'predict']

        dataPartitionExists = []
        # Check if a data partition exists, then convert to a pandas dataframe
        for i, data in enumerate([validateData, trainData, testData]):
            if data is not None:
                dataPartitionExists.append(i)
                if type(data) is np.ndarray:
                    try:
                        dataSets[i][columns] = data
                    except ValueError:
                        dataSets[i][columns] = data.transpose()
                elif type(data) is list:
                    dataSets[i][columns] = np.array(data).transpose()
                elif type(data) is pd.core.frame.DataFrame:
                    try:
                        dataSets[i][columns[0]] = data.iloc[:,0]
                        dataSets[i][columns[1]] = data.iloc[:,1]
                    except NameError:
                        dataSets[i] = pd.DataFrame(data=data.iloc[:,0]).rename(columns={data.columns[0]: columns[0]})
                        dataSets[i][columns[1]] = data.iloc[:,1]
                    
        if len(dataPartitionExists) == 0:
            print('No data was provided. Please provide the actual ' +
                    'and predicted values for at least one of the ' + 
                    'partitions (VALIDATE, TRAIN, or TEST).')
            raise ValueError

        nullLiftRow = list(range(1, 64))
        nullROCRow = list(range(1, 301))
        
        swatConn.loadactionset('percentile')
        
        for i in dataPartitionExists:
            swatConn.read_frame(dataSets[i][columns],
                                casout=dict(name='SCOREDVALUES', replace=True))
            swatConn.percentile.assess(table='SCOREDVALUES',
                                        inputs=[columns[1]],
                                        casout=dict(name='SCOREASSESS', replace=True),
                                        response=columns[0],
                                        event=str(targetValue))
            assessROC = swatConn.CASTable('SCOREASSESS_ROC').to_frame()
            assessLift = swatConn.CASTable('SCOREASSESS').to_frame()

            for j in range(100):
                rowNumber = (i*100) + j
                nullROCRow.remove(rowNumber + 1)
                nullJSONROCDict['data'][rowNumber]['dataMap']['_Event_'] = targetValue
                nullJSONROCDict['data'][rowNumber]['dataMap']['_TargetName_'] = targetName
                nullJSONROCDict['data'][rowNumber]['dataMap']['_Cutoff_'] = str(assessROC['_Cutoff_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_TP_'] = str(assessROC['_TP_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_FP_'] = str(assessROC['_FP_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_FN_'] = str(assessROC['_FN_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_TN_'] = str(assessROC['_TN_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_Sensitivity_'] = str(assessROC['_Sensitivity_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_Specificity_'] = str(assessROC['_Specificity_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_KS_'] = str(assessROC['_KS_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_KS2_'] = str(assessROC['_KS2_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_FHALF_'] = str(assessROC['_FHALF_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_FPR_'] = str(assessROC['_FPR_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_ACC_'] = str(assessROC['_ACC_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_FDR_'] = str(assessROC['_FDR_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_F1_'] = str(assessROC['_F1_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_C_'] = str(assessROC['_C_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_GINI_'] = str(assessROC['_GINI_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_GAMMA_'] = str(assessROC['_GAMMA_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_TAU_'] = str(assessROC['_TAU_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_MiscEvent_'] = str(assessROC['_MiscEvent_'][j])
                nullJSONROCDict['data'][rowNumber]['dataMap']['_OneMinusSpecificity_'] = str(1 - assessROC['_Specificity_'][j])
            
            for j in range(21):
                rowNumber = (i*21) + j
                nullLiftRow.remove(rowNumber + 1)
                nullJSONLiftDict['data'][rowNumber]['dataMap']['_Event_'] = str(targetValue)
                nullJSONLiftDict['data'][rowNumber]['dataMap']['_TargetName_'] = targetName
                if j != 0:
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_Depth_'] = str(assessLift['_Depth_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_Value_'] = str(assessLift['_Value_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_NObs_'] = str(assessLift['_NObs_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_NEvents_'] = str(assessLift['_NEvents_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_NEventsBest_'] = str(assessLift['_NEventsBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_Resp_'] = str(assessLift['_Resp_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_RespBest_'] = str(assessLift['_RespBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_Lift_'] = str(assessLift['_Lift_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_LiftBest_'] = str(assessLift['_LiftBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_CumResp_'] = str(assessLift['_CumResp_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_CumRespBest_'] = str(assessLift['_CumRespBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_CumLift_'] = str(assessLift['_CumLift_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_CumLiftBest_'] = str(assessLift['_CumLiftBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_PctResp_'] = str(assessLift['_PctResp_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_PctRespBest_'] = str(assessLift['_PctRespBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_CumPctResp_'] = str(assessLift['_CumPctResp_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_CumPctRespBest_'] = str(assessLift['_CumPctRespBest_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_Gain_'] = str(assessLift['_Gain_'][j-1])
                    nullJSONLiftDict['data'][rowNumber]['dataMap']['_GainBest_'] = str(assessLift['_GainBest_'][j-1])  
            
        # If not all partitions are present, clean up the dicts for compliant formatting        
        if len(dataPartitionExists) < 3:
            # Remove missing partitions from ROC and Lift dicts
            for index, row in reversed(list(enumerate(nullJSONLiftDict['data']))):
                if int(row['rowNumber']) in nullLiftRow:
                    nullJSONLiftDict['data'].pop(index)
            for index, row in reversed(list(enumerate(nullJSONROCDict['data']))):
                if int(row['rowNumber']) in nullROCRow:
                    nullJSONROCDict['data'].pop(index)
                    
            # Reassign the row number values to match what is left in each dict
            for i, _ in enumerate(nullJSONLiftDict['data']):
                nullJSONLiftDict['data'][i]['rowNumber'] = i + 1
            for i, _ in enumerate(nullJSONROCDict['data']):
                nullJSONROCDict['data'][i]['rowNumber'] = i + 1
        
        with open(Path(jPath) / 'dmcas_roc.json', 'w') as jFile:
            json.dump(nullJSONROCDict, jFile, indent=4) 
        print('{} was successfully written and saved to {}'.format('dmcas_roc.json', Path(jPath) / 'dmcas_roc.json'))
            
        with open(Path(jPath) / 'dmcas_lift.json', 'w') as jFile:
            json.dump(nullJSONLiftDict, jFile, indent=4)
        print('{} was successfully written and saved to {}'.format('dmcas_lift.json', Path(jPath) / 'dmcas_lift.json'))
                        
    def readJSONFile(self, path):
        '''
        Reads a JSON file from a given path.
        
        Parameters
        ----------
        path : str or pathlib Path
            Location of the JSON file to be opened.
            
        Returns
        -------
        json.load(jFile) : str
            String contents of JSON file.
        '''
        
        with open(path) as jFile:
            return(json.load(jFile))
            
    def formatParameter(self, paramName):
        '''
        Formats the parameter name to the JSON standard (_<>_). No changes are
        applied if the string is already formatted correctly.
        
        Parameters
        ---------------
        paramName : string
            Name of the parameter.
            
        Returns
        ---------------
        paramName : string
            Name of the parameter.
        '''
        
        if not (paramName.startswith('_') and paramName.endswith('_')):
            if not paramName.startswith('_'):
                paramName = '_' + paramName
            if not paramName.endswith('_'):
                paramName = paramName + '_'
        
        return paramName
            
    def convertDataRole(self, dataRole):
        '''
        Converts the data role identifier from string to int or int to string. JSON
        file descriptors require the string, int, and formatted int. If the
        provided data role is not valid, defaults to TRAIN (1).
        
        Parameters
        ---------------
        dataRole : string or int
            Identifier of the data set's role; either TRAIN, TEST, or VALIDATE,
            which correspond to 1, 2, or 3.
            
        Returns
        ---------------
        conversion : int or string
            Converted data role identifier.
        '''
        
        if type(dataRole) is int or type(dataRole) is float:
            dataRole = int(dataRole)
            if dataRole == 1:
                conversion = 'TRAIN'
            elif dataRole == 2:
                conversion = 'TEST'
            elif dataRole == 3:
                conversion = 'VALIDATE'
            else:
                conversion = 'TRAIN'
        elif type(dataRole) is str:
            if dataRole == 'TRAIN':
                conversion = 1
            elif dataRole == 'TEST':
                conversion = 2    
            elif dataRole == 'VALIDATE':
                conversion = 3
            else:
                conversion = 1
        else:
            conversion = 1
                
        return conversion


#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Enables publishing objects such as models to various destinations."""

import re

from .model_repository import ModelRepository
from .service import Service

class ModelPublish(Service):
    """Enables publishing objects such as models to various destinations.

    The service provides methods for managing publish destinations like CAS,
    Hadoop, Teradata, or SAS Micro Analytic Service
    """

    _SERVICE_ROOT = '/modelPublish'
    _model_repository = ModelRepository()

    @staticmethod
    def _publish_name(name):
        """Create a valid module name from an input string.

        Model Publishing only permits names that adhere to the following
        restrictions:
            - Name must start with a letter or an underscore.
            - Name may not contain any spaces or special characters other than
              underscore.

        Parameters
        ----------
        name : str

        Returns
        -------
        str

        """
        # Remove all non-word characters
        name = re.sub(r'[^\w]', '', str(name))

        # Add a leading underscore if name starts with a digit
        if re.match(r'\d', name):
            name = '_' + name

        return name

    @classmethod
    def list_models(cls):
        return cls.get('/models').get('items', [])

    list_destinations, get_destination, update_destination, \
    _ = Service._crud_funcs('/destinations',
                                                 'destination')

    # @sasctl_command('delete')
    @classmethod
    def delete_destination(cls, item):
        """Delete a destination instance.

        Parameters
        ----------
        item : str or dict
            Name or dictionary representation of publishing destination.

        Returns
        -------
        None

        """

        # NOTE:  Explicitly defined since service requires destination NAME not
        # ID which is standard.
        item_name = str(item)

        # Try to find the item if the id can't be found
        if not (isinstance(item, dict) and 'id' in item):
            item = cls.get_destination(item)
            if item is None:
                cls.log.info("Object '%s' not found.  Skipping delete.",
                             item_name)
                return

        if isinstance(item, dict) and 'name' in item:
            item = item['name']

        return cls.delete('/destinations/{name}'.format(name=item))

    @classmethod
    def publish_model(cls,
                      model,
                      destination,
                      name=None,
                      code=None,
                      notes=None):
        """Publish a model to an existing publishing destination.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model.
        destination : str or dict
            The name or id of the publishing destination, or a dictionary
            representation of the destination
        name : str, optional
            Name of the published model.  Defaults to the model name.
        code : str, optional
            The code to be published.
        notes

        Returns
        -------

        """
        code_types = {
            'ds2package': 'ds2',
            'datastep': 'datastep',
            '': ''
        }

        model = cls._model_repository.get_model(model)
        model_uri = cls._model_repository.get_model_link(model, 'cls')

        # Get score code from registry if no code specified
        if code is None:
            code_link = cls._model_repository.get_model_link(model,
                                                             'scoreCode',
                                                             True)
            if code_link:
                code = cls.get(code_link['href'])

        request = dict(
            name=name or model.get('name'),
            notes=notes,
            destinationName=destination,
        )

        modelContents = {
            'modelName': model.get('name'),
            'modelId': model.get('id'),
            'sourceUri': model_uri.get('href'),
            'publishLevel': 'model',        # ?? What are the options?
            'codeType': code_types[model.get('scoreCodeType', '').lower()],
            'codeUri': '',          # ??  Not needed if code is specified?
            'code': code
        }

        request['modelContents'] = [modelContents]
        return cls.post('/models', json=request, headers={
            'Content-Type':
                'application/vnd.sas.models.publishing.request+json'})

    @classmethod
    def create_cas_destination(cls, name,
                               library,
                               table,
                               server=None,
                               description=None):
        """Define a new CAS publishing destination.

        Parameters
        ----------
        name : str
            Name of the publishing destination.
        library : str
            The CAS library in which `table` will be stored.
        table : str
            Name of the CAS table in which models will be stored.
        server : str, optional
            Name of the CAS server.  Defaults to 'cas-shared-default'.
        description : str, optional
            Description of the publishing destination.

        Returns
        -------
        RestObj

        """
        server = server or 'cas-shared-default'

        return cls.create_destination(name, cas_server=server,
                                      cas_library=library, cas_table=table,
                                      type_='cas', description=description)

    @classmethod
    def create_mas_destination(cls, name, uri, description=None):
        """Define a new Micro Analytic Server (MAS) publishing destination.

        Parameters
        ----------
        name : str
            Name of the publishing destination.
        uri : str
            The base URI that contains the host, the protocol, and optionally
            the port, which addresses the remote SAS Micro Analytic Service to
            use.  Example: http://spam.com
        description : str, optional
            Description of the publishing destination.

        Returns
        -------
        RestObj

        """
        return cls.create_destination(name, mas_uri=uri, type_='mas',
                                      description=description)

    @classmethod
    def create_destination(cls,
                           name,
                           type_,
                           cas_server=None,
                           cas_library=None,
                           cas_table=None,
                           description=None,
                           mas_uri=None,
                           hdfs_dir=None,
                           conf_dir=None,
                           user=None,
                           database_library=None):
        """Define a new publishing destination.

        Parameters
        ----------
        name : str
            Name of the publishing destination.
        type_ : {'cas', 'mas', 'hadoop', 'teradata'}
            Type of publishing definition being created
        cas_server : str, optional
            Name of the CAS server.  Defaults to 'cas-shared-default'.
            Required if `type_` is 'cas', otherwise ignored.
        cas_library : str, optional
            The CAS library in which `cas_table` will be stored.  Required if
            `type_` is 'cas', otherwise ignored.
        cas_table : str, optional
            Name of the CAS table in which models will be stored.  Required
            if `type_` is 'cas', otherwise ignored.
        description : str, optional
            Description of the publishing destination.
        mas_uri : str, optional
            Required if `type_` is 'mas', otherwise ignored.
        hdfs_dir : str, optional
            Required if `type_` is 'hadoop', otherwise ignored.
        conf_dir : str, optional
            Required if `type_` is 'hadoop', otherwise ignored.
        user : str, optional
            Required if `type_` is 'hadoop', otherwise ignored.
        database_library : str, optional
            Required if `type_` is 'teradata', otherwise ignored.

        Returns
        -------
        RestObj

        """
        type_ = str(type_).lower()
        if type_ not in ('cas', 'microanalyticservice', 'mas', 'teradata',
                         'hadoop'):
            raise ValueError("Unrecognized destination type '%s' specified."
                             % type_)

        # As of Viya 3.4 capitalization matters.
        if type_ in ('microanalyticservice', 'mas'):
            type_ = 'microAnalyticService'

        request = {'name': str(name),
                   'destinationType': type_,
                   'casServerName': cas_server,
                   'casLibrary': cas_library,
                   'description': description,
                   'destinationTable': cas_table,
                   'databaseCasLibrary': database_library,
                   'user': user,
                   'hdfsDirectory': hdfs_dir,
                   'configurationDirectory': conf_dir,
                   'masUri': mas_uri
                   }

        drop_list = {
            'cas':
                ('databaseCasLibrary', 'user', 'hdfsDirectory', 'masUri',
                 'configurationDirectory'),
            'microAnalyticService':
                ('casServerName', 'casLibrary', 'destinationTable', 'user',
                 'databaseCasLibrary', 'hdfsDirectory',
                 'configurationDirectory'),
            'hadoop':
                ('casServerName', 'casLibrary', 'destinationTable',
                 'databaseCasLibrary', 'masUri'),
            'teradata':
                ('casServerName', 'casLibrary', 'destinationTable', 'user',
                 'hdfsDirectory', 'masUri', 'configurationDirectory')
        }

        for k in drop_list[request['destinationType']]:
            request.pop(k, None)

        return cls.post('/destinations', json=request, headers={
            'Content-Type':
                'application/vnd.sas.models.publishing.destination+json'})



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import io
import json
import os
import shutil
import tempfile
import uuid
import zipfile


try:
    import swat
except ImportError:
    swat = None


def create_package(table, input=None):
    """Create an importable model package from a CAS table.

    Parameters
    ----------
    table : swat.CASTable
        The CAS table containing an ASTORE or score code.
    input : DataFrame, type, list of type, or dict of str: type, optional
        The expected type for each input value of the target function.
        Can be omitted if target function includes type hints.  If a DataFrame
        is provided, the columns will be inspected to determine type information.
        If a single type is provided, all columns will be assumed to be that type,
        otherwise a list of column types or a dictionary of column_name: type
        may be provided.

    Returns
    -------
    BytesIO
        A byte stream representing a ZIP archive which can be imported.

    See Also
    --------
    :meth:`model_repository.import_model_from_zip <.ModelRepository.import_model_from_zip>`

    """
    if swat is None:
        raise RuntimeError("The 'swat' package is required to work with SAS models.")

    if not isinstance(table, swat.CASTable):
        raise ValueError(
            "Parameter 'table' should be an instance of '%r' but "
            "received '%r'." % (swat.CASTable, table)
        )

    if 'DataStepSrc' in table.columns:
        # Input only passed to datastep
        return create_package_from_datastep(table, input=input)
    return create_package_from_astore(table)


def create_package_from_datastep(table, input=None):
    """Create an importable model package from a score code table.

    Parameters
    ----------
    table : swat.CASTable
        The CAS table containing the score code.
    input : DataFrame, type, list of type, or dict of str: type, optional
        The expected type for each input value of the target function.
        Can be omitted if target function includes type hints.  If a DataFrame
        is provided, the columns will be inspected to determine type information.
        If a single type is provided, all columns will be assumed to be that type,
        otherwise a list of column types or a dictionary of column_name: type
        may be provided.

    Returns
    -------
    BytesIO
        A byte stream representing a ZIP archive which can be imported.

    See Also
    --------
    :meth:`model_repository.import_model_from_zip <.ModelRepository.import_model_from_zip>`

    """
    dscode = table.to_frame().loc[0, 'DataStepSrc']

    # Extract inputs if provided
    input_vars = []
    # Workaround because sasdataframe does not like to be check if exist
    if str(input) != "None":
        from .pymas.python import ds2_variables

        variables = None
        if hasattr(input, 'columns'):
            # Assuming input is a DataFrame representing model inputs.  Use to
            # get input variables
            variables = ds2_variables(input)
        elif isinstance(input, dict):
            variables = ds2_variables(input)
        if variables:
            input_vars = [v.as_model_metadata() for v in variables if not v.out]

    # Find outputs from ds code
    output_vars = []
    for sasline in dscode.split('\n'):
        if sasline.strip().startswith('label'):
            output_var = {}
            for tmp in sasline.split('='):
                if 'label' in tmp:
                    ovarname = tmp.split('label')[1].strip()
                    output_var.update({"name": ovarname})
                    # Determine type of variable is decimal or string
                    if "length " + ovarname in dscode:
                        sastype = (
                            dscode.split("length " + ovarname)[1].split(';')[0].strip()
                        )
                        if "$" in sastype:
                            output_var.update({"type": "string"})
                            output_var.update({"length": sastype.split("$")[1]})
                        else:
                            output_var.update({"type": "decimal"})
                            output_var.update({"length": sastype})
                    else:
                        # If no length for variable, default is decimal, 8
                        output_var.update({"type": "decimal"})
                        output_var.update({"length": 8})
                else:
                    output_var.update(
                        {"description": tmp.split(';')[0].strip().strip("'")}
                    )
            output_vars.append(output_var)

    file_metadata = [{'role': 'score', 'name': 'dmcas_scorecode.sas'}]

    zip_file = _build_zip_from_files(
        {
            'fileMetadata.json': file_metadata,
            'dmcas_scorecode.sas': dscode,
            'ModelProperties.json': {"scoreCodeType": "dataStep"},
            'outputVar.json': output_vars,
            'inputVar.json': input_vars,
        }
    )

    return zip_file


def create_package_from_astore(table):
    """Create an importable model package from an ASTORE.

    Parameters
    ----------
    table : swat.CASTable
        The CAS table containing the ASTORE.

    Returns
    -------
    BytesIO
        A byte stream representing a ZIP archive which can be imported.

    See Also
    --------
    :meth:`model_repository.import_model_from_zip <.ModelRepository.import_model_from_zip>`

    """
    files = create_files_from_astore(table)

    return _build_zip_from_files(files)


def create_files_from_astore(table):
    """Generate files for importing a model from an ASTORE.

    Parameters
    ----------
    table : swat.CASTable
        The CAS table containing the ASTORE.

    Returns
    -------
    dict
        Dictionary of filename: content pairs.

    """
    if swat is None:
        raise RuntimeError(
            "The 'swat' package is required to work with " "ASTORE models."
        )

    if not isinstance(table, swat.CASTable):
        raise ValueError(
            "Parameter 'table' should be an instance of '%r' but "
            "received '%r'." % (swat.CASTable, table)
        )

    sess = table.session.get_connection()
    sess.loadactionset('astore')

    result = sess.astore.describe(rstore=table, epcode=True)

    # Model Manager expects a 0-byte ASTORE file.  Will retrieve actual ASTORE
    # from CAS during model publish.
    astore = bytes()

    # Raise error if describe action fails
    if result.status_code != 0:
        raise RuntimeError(result)

    astore_key = result.Key.Key[0].strip()

    # Remove "Keep" sas code from CAS/EP code so full table plus output are
    # returned. This is so the MM performance charts and test work.
    keepstart = result.epcode.find("Keep")
    keepend = result.epcode.find(";", keepstart)
    ep_ds2 = result.epcode[0:keepstart] + result.epcode[keepend + 1 :]

    package_ds2 = _generate_package_code(result)
    model_properties = _get_model_properties(result)
    input_vars = [
        get_variable_properties(var) for var in result.InputVariables.itertuples()
    ]
    input_vars = [v for v in input_vars if v.get('role', '').upper() == 'INPUT']
    output_vars = [
        get_variable_properties(var) for var in result.OutputVariables.itertuples()
    ]
    astore_filename = '_' + uuid.uuid4().hex[:25].upper()

    # Copy the ASTORE table to the ModelStore.
    # Raise an error if the action fails
    with swat.options(exception_on_severity=2):
        table.save(name=astore_filename, caslib='ModelStore', replace=True)

    file_metadata = [
        {'role': 'analyticStore', 'name': ''},
        {'role': 'score', 'name': 'dmcas_epscorecode.sas'},
    ]

    astore_metadata = [
        {
            'name': astore_filename,
            'caslib': 'ModelStore',
            'uri': '/dataTables/dataSources/cas~fs~cas-shared-default~fs~ModelStore/tables/{}'.format(
                astore_filename
            ),
            'key': astore_key,
        }
    ]

    return {
        'dmcas_packagescorecode.sas': '\n'.join(package_ds2),
        'dmcas_epscorecode.sas': ep_ds2,
        astore_filename: astore,
        'ModelProperties.json': model_properties,
        'fileMetadata.json': file_metadata,
        'AstoreMetadata.json': astore_metadata,
        'inputVar.json': input_vars,
        'outputVar.json': output_vars,
    }


def _build_zip_from_files(files):
    """Create a ZIP file containing the provided files.

    Parameters
    ----------
    files : dict
        Dictionary of filename: content to be added to the .zip file.

    Returns
    -------
    BytesIO
        Byte stream representation of the .zip file.

    """
    try:
        # Create a temp folder
        folder = tempfile.mkdtemp()

        for k, v in files.items():
            filename = os.path.join(folder, k)

            # Write JSON file
            if os.path.splitext(k)[-1].lower() == '.json':
                with open(filename, 'w') as f:
                    json.dump(v, f, indent=1)
            else:
                mode = 'wb' if isinstance(v, bytes) else 'w'

                with open(filename, mode) as f:
                    f.write(v)

        files = os.listdir(folder)

        with zipfile.ZipFile(os.path.join(folder, 'model.zip'), 'w') as z:
            for file in files:
                z.write(os.path.join(folder, file), file)

        # Need to return the ZIP file data but also need to ensure the
        # directory is cleaned up.
        # Read the bytes from disk and return an in memory "file".
        with open(os.path.join(folder, 'model.zip'), 'rb') as z:
            return io.BytesIO(z.read())
    finally:
        shutil.rmtree(folder)


def get_variable_properties(var):
    type_mapping = {'interval': '', 'num': 'decimal', 'character': 'string'}

    meta = {'name': var.Name.strip(), 'length': int(var.Length)}

    # Input variable table has Type & RawType columns, but RawType aligns with Type column from Output variable table.
    if hasattr(var, 'RawType'):
        meta['type'] = type_mapping[var.RawType.strip().lower()]
    else:
        meta['type'] = type_mapping[var.Type.strip().lower()]

    if hasattr(var, 'Role'):
        meta['role'] = var.Role.strip().upper()

    return meta


def _get_model_properties(result):
    properties = {
        "custom properties": [],
        "externalUrl": "",
        "trainTable": "",
        "trainCodeType": "",
        "description": "",
        "tool": 'SAS Visual Data Mining and Machine Learning',
        "toolVersion": "",
        "targetVariable": '',
        "scoreCodeType": "ds2MultiType",
        "externalModelId": "",
        "function": '',
        "eventProbVar": "",
        "modeler": "",
        "name": "",
        "targetEvent": "",
        "targetLevel": "",
        "algorithm": '',
    }

    algorithm = result.Description[result.Description.Attribute == 'Analytic Engine']
    if algorithm.size > 0:
        algorithm = str(algorithm.Value.iloc[0]).lower()
    else:
        algorithm = None

    def classification_target(r):
        target = r.OutputVariables.Name.str.startswith('I_')
        target = r.OutputVariables.Name[target].iloc[0]
        return target.replace('I_', '', 1)

    def regression_target(r):
        target = r.OutputVariables.Name.str.startswith('P_')
        target = r.OutputVariables.Name[target].iloc[0]
        return target.replace('P_', '', 1)

    if algorithm == 'glm':
        properties['algorithm'] = 'Linear regression'
        properties['tool'] = 'SAS Visual Analytics'
        properties['function'] = 'prediction'
        properties['targetVariable'] = regression_target(result)

    elif algorithm == 'logistic':
        properties['algorithm'] = 'Logistic regression'
        properties['tool'] = 'SAS Visual Analytics'
        properties['function'] = 'classification'
        properties['targetVariable'] = classification_target(result)

    elif algorithm == 'forest':
        properties['algorithm'] = 'Random forest'

        if 'Classification' in result.InputVariables.Type.values:
            properties['function'] = 'classification'
            properties['targetVariable'] = classification_target(result)
        else:
            properties['function'] = 'prediction'
            properties['targetVariable'] = regression_target(result)

    elif algorithm == 'gradboost':
        properties['algorithm'] = 'Gradient boosting'

        if 'Classification' in result.InputVariables.Type.values:
            properties['function'] = 'classification'
            properties['targetVariable'] = classification_target(result)

            if result.OutputVariables.Name.str.startswith('P_').sum() == 2:
                properties['targetLevel'] = 'binary'
        else:
            properties['function'] = 'prediction'
            properties['targetVariable'] = regression_target(result)

    elif algorithm == 'svmachine':
        properties['algorithm'] = 'Support vector machine'

        if 'Classification' in result.InputVariables.Type.values:
            properties['function'] = 'classification'
            properties['targetVariable'] = classification_target(result)
            properties['targetLevel'] = 'binary'
        else:
            properties['function'] = 'prediction'
            properties['targetVariable'] = regression_target(result)

    elif algorithm == 'bnet':
        properties['algorithm'] = 'Bayesian network'
        properties['function'] = 'classification'
        properties['targetVariable'] = classification_target(result)

        if result.OutputVariables.Name.str.startswith('P_').sum() == 2:
            properties['targetLevel'] = 'binary'

    else:
        properties['tool'] = ''

    return properties


def _generate_package_code(result):
    """Generates package-style DS2 code from EP-style DS2 code."""

    id_ = '_' + uuid.uuid4().hex  # Random ID for package
    key = result.Key.Key[0]

    header = (
        'package ds2score / overwrite=yes;',
        '    dcl package score {}();'.format(id_),
    )

    dcl_lines = []
    for line in result.epcode.split('\n'):
        # Ignore the package declaration since it will be redefined
        if line.strip().startswith('dcl ') and not line.strip().startswith(
            'dcl package '
        ):
            dcl_lines.append(line)

    init_method = (
        '    varlist allvars [_all_];',
        ' ',
        '    method init();',
        "       {}.setvars(allvars);".format(id_),
        "       {}.setkey(n'{}');".format(id_, key),
        '    end;',
    )

    def extract_type(var, out=False):
        # Find the matching variable declarations and extract the type
        var = str(var).strip()
        x = [x for x in dcl_lines if ' "{}"'.format(var) in x][0]
        x = x.replace('dcl ', '').strip().split(' ')[0]

        # Remove the length component from output variables to prevent
        # compilation warning which prevents publishing to MAS
        if out and '(' in x:
            x = x[: x.find('(')]
        return x

    variables = []
    # Despite being call "InputVariables" at least some ASTORE models
    # include the target variable in the list
    for _, row in result.InputVariables.iterrows():
        if 'Role' in row and row['Role'].lower() != 'target':
            name = row['Name']
            variables.append('       %s "%s"' % (extract_type(name), name))
    variables += [
        '       IN_OUT {} "{}"'.format(extract_type(var, out=True), var)
        for var in result.OutputVariables.Name
    ]

    score_method = ('    method score(', ',\n'.join(variables), '   );')
    score_method += tuple(
        '       this."{var}" = "{var}";'.format(var=v)
        for v in result.InputVariables.Name
    )
    score_method += (' ', '       {}.scorerecord();'.format(id_), ' ')
    score_method += tuple(
        '       "{var}" = this."{var}";'.format(var=v)
        for v in result.OutputVariables.Name
    )

    footer = ('    end;', 'endpackage;')

    return header + tuple(dcl_lines) + init_method + score_method + footer



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Commonly used tasks in the analytics life cycle."""

import json
import logging
import math
import pickle  # skipcq BAN-B301
import os
import re
import sys
import warnings

from urllib.error import HTTPError

from . import utils
from .core import RestObj, current_session, get, get_link, request_link
from .exceptions import AuthorizationError
from .services import model_management as mm
from .services import model_publish as mp
from .services import model_repository as mr
from .utils.pymas import from_pickle
from .utils.misc import installed_packages


logger = logging.getLogger(__name__)

# As of Viya 3.4 model registration fails if character fields are longer
# than 1024 characters
_DESC_MAXLEN = 1024

# As of Viya 3.4 model registration fails if user-defined properties are
# longer than 512 characters.
_PROP_VALUE_MAXLEN = 512
_PROP_NAME_MAXLEN = 60


def _property(k, v):
    return {'name': str(k)[:_PROP_NAME_MAXLEN], 'value': str(v)[:_PROP_VALUE_MAXLEN]}


def _sklearn_to_dict(model):
    # Convert Scikit-learn values to built-in Model Manager values
    mappings = {
        'LogisticRegression': 'Logistic regression',
        'LinearRegression': 'Linear regression',
        'SVC': 'Support vector machine',
        'GradientBoostingClassifier': 'Gradient boosting',
        'GradientBoostingRegressor': 'Gradient boosting',
        'XGBClassifier': 'Gradient boosting',
        'XGBRegressor': 'Gradient boosting',
        'RandomForestClassifier': 'Forest',
        'DecisionTreeClassifier': 'Decision tree',
        'DecisionTreeRegressor': 'Decision tree',
        'classifier': 'classification',
        'regressor': 'prediction',
    }

    if hasattr(model, '_final_estimator'):
        estimator = model._final_estimator
    else:
        estimator = model
    estimator = type(estimator).__name__

    # Standardize algorithm names
    algorithm = mappings.get(estimator, estimator)

    # Standardize regression/classification terms
    analytic_function = mappings.get(model._estimator_type, model._estimator_type)

    if analytic_function == 'classification' and 'logistic' in algorithm.lower():
        target_level = 'Binary'
    elif analytic_function == 'prediction' and (
        'regressor' in estimator.lower() or 'regression' in algorithm.lower()
    ):
        target_level = 'Interval'
    else:
        target_level = None

    # Can tell if multi-class .multi_class
    result = dict(
        description=str(model)[:_DESC_MAXLEN],
        algorithm=algorithm,
        scoreCodeType='ds2MultiType',
        trainCodeType='Python',
        targetLevel=target_level,
        function=analytic_function,
        tool='Python %s.%s' % (sys.version_info.major, sys.version_info.minor),
        properties=[_property(k, v) for k, v in model.get_params().items()],
    )

    return result


def _create_project(project_name, model, repo, input_vars=None, output_vars=None):
    """Creates a project based on the model specifications.

    Parameters
    ----------
    project_name : str
        Name of the project to be created
    model : dict
        Model information
    repo : str or dict
        Repository in which to create the project
    input_vars : list
        Input variables formatted as {'name': 'varname'}
    output_vars
        Output variables formatted as {'name': 'varname'}

    Returns
    -------
    RestObj
        The created project
    """
    properties = {k: model[k] for k in model if k in ('function', 'targetLevel')}

    function = model.get('function', '').lower()
    algorithm = model.get('algorithm', '').lower()

    # Get input & output variable lists
    # Note: copying lists to avoid altering original
    input_vars = input_vars or model.get('inputVariables', [])
    output_vars = output_vars or model.get('outputVariables', [])[:]
    input_vars = input_vars[:]
    output_vars = output_vars[:]

    # Set prediction or eventProbabilityVariable
    if output_vars:
        if function == 'prediction':
            properties['predictionVariable'] = output_vars[0]['name']
        else:
            properties['eventProbabilityVariable'] = output_vars[0]['name']

    # Set targetLevel
    if properties.get('targetLevel') is None:
        if function == 'classification' and 'logistic' in algorithm:
            properties['targetLevel'] = 'Binary'
        elif function == 'prediction' and 'regression' in algorithm:
            properties['targetLevel'] = 'Interval'
        else:
            properties['targetLevel'] = None

    project = mr.create_project(
        project_name, repo, variables=input_vars + output_vars, **properties
    )

    # As of Viya 3.4 the 'predictionVariable' and 'eventProbabilityVariable'
    # parameters are not set during project creation.  Update the project if
    # necessary.
    needs_update = False
    for p in ('predictionVariable', 'eventProbabilityVariable'):
        if project.get(p) != properties.get(p):
            project[p] = properties.get(p)
            needs_update = True

    if needs_update:
        project = mr.update_project(project)

    return project


def register_model(
    model,
    name,
    project,
    repository=None,
    input=None,
    version=None,
    files=None,
    force=False,
    record_packages=True,
):
    """Register a model in the model repository.

    Parameters
    ----------
    model : swat.CASTable or sklearn.BaseEstimator
        The model to register.  If an instance of ``swat.CASTable`` the table
        is assumed to hold an ASTORE, which will be downloaded and used to
        construct the model to register.  If a scikit-learn estimator, the
        model will be pickled and uploaded to the registry and score code will
        be generated for publishing the model to MAS.
    name : str
        Designated name for the model in the repository.
    project : str or dict
        The name or id of the project, or a dictionary representation of
        the project.
    repository : str or dict, optional
        The name or id of the repository, or a dictionary representation of
        the repository.  If omitted, the default repository will be used.
    input : DataFrame, type, list of type, or dict of str: type, optional
        The expected type for each input value of the target function.
        Can be omitted if target function includes type hints.  If a DataFrame
        is provided, the columns will be inspected to determine type
        information.  If a single type is provided, all columns will be assumed
        to be that type, otherwise a list of column types or a dictionary of
        column_name: type may be provided.
    version : {'new', 'latest', int}, optional
        Version number of the project in which the model should be created.
        Defaults to 'new'.
    files : list
        A list of dictionaries of the form
        {'name': filename, 'file': filecontent}.
        An optional 'role' key is supported for designating a file as score
        code, astore, etc.
    force : bool, optional
        Create dependencies such as projects and repositories if they do not
        already exist.
    record_packages : bool, optional
        Capture Python packages registered in the environment.  Defaults to
        True.  Ignored if `model` is not a Python object.

    Returns
    -------
    model : RestObj
        The newly registered model as an instance of ``RestObj``

    Notes
    -----
    If the specified model is a CAS table the model data and metadata will be
    written to a temporary zip file and then imported using
    model_repository.import_model_from_zip.

    If the specified model is from the Scikit-Learn package, the model will be
    created using model_repository.create_model and any additional files will
    be uploaded as content.

    .. versionchanged:: v1.3
        Create requirements.txt with installed packages.

    .. versionchanged:: v1.4.5
        Added `record_packages` parameter.

    """
    # TODO: Create new version if model already exists

    # If version not specified, default to creating a new version
    version = version or 'new'

    files = files or []

    # Find the project if it already exists
    p = mr.get_project(project) if project is not None else None

    # Do we need to create the project first?
    create_project = bool(p is None and force is True)

    if p is None and not create_project:
        raise ValueError("Project '{}' not found".format(project))

    # Use default repository if not specified
    try:
        if repository is None:
            repo_obj = mr.default_repository()
        else:
            repo_obj = mr.get_repository(repository)
    except HTTPError as e:
        if e.code == 403:
            raise AuthorizationError(
                'Unable to register model.  User account '
                'does not have read permissions for the '
                '/modelRepository/repositories/ URL. '
                'Please contact your SAS Viya '
                'administrator.'
            )
        raise e

    # Unable to find or create the repo.
    if repo_obj is None and repository is None:
        raise ValueError("Unable to find a default repository")

    if repo_obj is None:
        raise ValueError("Unable to find repository '{}'".format(repository))

    # If model is a CASTable then assume it holds an ASTORE model.
    # Import these via a ZIP file.
    if 'swat.cas.table.CASTable' in str(type(model)):
        zipfile = utils.create_package(model, input=input)

        if create_project:
            outvar = []
            invar = []
            import zipfile as zp
            import copy

            zipfilecopy = copy.deepcopy(zipfile)
            tmpzip = zp.ZipFile(zipfilecopy)
            if "outputVar.json" in tmpzip.namelist():
                outvar = json.loads(
                    tmpzip.read("outputVar.json").decode('utf=8')
                )  # added decode for 3.5 and older
                for tmp in outvar:
                    tmp.update({'role': 'output'})
            if "inputVar.json" in tmpzip.namelist():
                invar = json.loads(
                    tmpzip.read("inputVar.json").decode('utf-8')
                )  # added decode for 3.5 and older
                for tmp in invar:
                    if tmp['role'] != 'input':
                        tmp['role'] = 'input'

            if 'ModelProperties.json' in tmpzip.namelist():
                model_props = json.loads(
                    tmpzip.read('ModelProperties.json').decode('utf-8')
                )
            else:
                model_props = {}

            project = _create_project(project, model_props, repo_obj, invar, outvar)

        model = mr.import_model_from_zip(name, project, zipfile, version=version)
        return model

    # If the model is an scikit-learn model, generate the model dictionary
    # from it and pickle the model for storage
    if all(hasattr(model, attr) for attr in ['_estimator_type', 'get_params']):
        # Pickle the model so we can store it
        model_pkl = pickle.dumps(model)
        files.append({'name': 'model.pkl', 'file': model_pkl, 'role': 'Python Pickle'})

        target_funcs = [f for f in ('predict', 'predict_proba') if hasattr(model, f)]

        # Extract model properties
        model = _sklearn_to_dict(model)
        model['name'] = name

        # Get package versions in environment
        packages = installed_packages()
        if record_packages and packages is not None:
            model.setdefault('properties', [])

            # Define a custom property to capture each package version
            # NOTE: some packages may not conform to the 'name==version' format
            #  expected here (e.g those installed with pip install -e). Such
            #  packages also generally contain characters that are not allowed
            # in custom properties, so they are excluded here.
            for p in packages:
                if '==' in p:
                    n, v = p.split('==')
                    model['properties'].append(_property('env_%s' % n, v))

            # Generate and upload a requirements.txt file
            files.append({'name': 'requirements.txt', 'file': '\n'.join(packages)})

        # Generate PyMAS wrapper
        try:
            mas_module = from_pickle(
                model_pkl, target_funcs, input_types=input, array_input=True
            )

            # Include score code files from ESP and MAS
            files.append(
                {
                    'name': 'dmcas_packagescorecode.sas',
                    'file': mas_module.score_code(),
                    'role': 'Score Code',
                }
            )
            files.append(
                {
                    'name': 'dmcas_epscorecode.sas',
                    'file': mas_module.score_code(dest='CAS'),
                    'role': 'score',
                }
            )
            files.append(
                {
                    'name': 'python_wrapper.py',
                    'file': mas_module.score_code(dest='Python'),
                }
            )

            model['inputVariables'] = [
                var.as_model_metadata() for var in mas_module.variables if not var.out
            ]

            model['outputVariables'] = [
                var.as_model_metadata() for var in mas_module.variables if var.out
            ]
        except ValueError:
            # PyMAS creation failed, most likely because input data wasn't
            # provided
            logger.exception('Unable to inspect model %s', model)

            warnings.warn(
                'Unable to determine input/output variables. '
                ' Model variables will not be specified and some '
                'model functionality may not be available.'
            )
    else:
        # Otherwise, the model better be a dictionary of metadata
        if not isinstance(model, dict):
            raise TypeError(
                "Expected an instance of '%r' but received '%r'." % ({}, model)
            )

    if create_project:
        project = _create_project(project, model, repo_obj)

    # If replacing an existing version, make sure the model version exists
    if str(version).lower() != 'new':
        # Update an existing model with new files
        model_obj = mr.get_model(name)
        if model_obj is None:
            raise ValueError(
                "Unable to update version '%s' of model '%s.  "
                "Model not found." % (version, name)
            )
        model = mr.create_model_version(name)
        mr.delete_model_contents(model)
    else:
        # Assume new model to create
        model = mr.create_model(model, project)

    if not isinstance(model, RestObj):
        raise TypeError(
            "Model should be an instance of '%r' but received '%r' "
            "instead." % (RestObj, model)
        )

    # Upload any additional files
    for file in files:
        if isinstance(file, dict):
            mr.add_model_content(model, **file)
        else:
            mr.add_model_content(model, file)

    return model


def publish_model(
    model, destination, code=None, max_retries=60, replace=False, **kwargs
):
    """Publish a model to a configured publishing destination.

    Parameters
    ----------
    model : str or dict
        The name or id of the model, or a dictionary representation of
        the model.
    destination : str
    code : optional
    max_retries : int, optional
    replace : bool, optional
        Whether to overwrite the model if it already exists in
        the `destination`
    kwargs : optional
        additional arguments will be passed to the underlying publish
        functions.

    Returns
    -------
    RestObj
        The published model

    Notes
    -----
    If no code is specified, the model is assumed to be already registered in
    the model repository and Model Manager's publishing functionality will be
    used.

    Otherwise, the model publishing API will be used.

    See Also
    --------
    :meth:`model_management.publish_model <.ModelManagement.publish_model>`
    :meth:`model_publish.publish_model <.ModelPublish.publish_model>`


    .. versionchanged:: 1.1.0
       Added `replace` option.

    """

    def submit_request():
        # Submit a publishing request
        if code is None:
            dest_obj = mp.get_destination(destination)

            if dest_obj and dest_obj.destinationType == "cas":
                publish_req = mm.publish_model(
                    model, destination, force=replace, reload_model_table=True
                )
            else:
                publish_req = mm.publish_model(model, destination, force=replace)
        else:
            publish_req = mp.publish_model(model, destination, code=code, **kwargs)

        # A successfully submitted request doesn't mean a successfully
        # published model.  Response for publish request includes link to
        # check publish log
        result = mr._monitor_job(publish_req, max_retries=max_retries)
        return result

    # Submit and wait for status
    job = submit_request()

    # If model was successfully published and it isn't a MAS module, we're done
    if (
        job.state.lower() == 'completed'
        and job.destination.destinationType != 'microAnalyticService'
    ):
        return request_link(job, 'self')

    # If MAS publish failed and replace=True, attempt to delete the module
    # and republish
    if (
        job.state.lower() == 'failed'
        and replace
        and job.destination.destinationType == 'microAnalyticService'
    ):
        from .services import microanalytic_score as mas

        mas.delete_module(job.publishName)

        # Resubmit the request
        job = submit_request()

    # Raise exception if still failing
    if job.state.lower() == 'failed':
        log = request_link(job, 'publishingLog')
        raise RuntimeError("Failed to publish model '%s': %s" % (model, log.log))

    # Raise exception if unknown status received
    if job.state.lower() != 'completed':
        raise RuntimeError(
            "Model publishing job in an unknown state: '%s'" % job.state.lower()
        )

    log = request_link(job, 'publishingLog')
    msg = log.get('log').lstrip('SUCCESS===')

    # As of Viya 3.4 MAS converts module names to lower case.
    # Since we can't rely on the request module name being preserved, try to
    # parse the URL out of the response so we can retrieve the created module.
    module_url = _parse_module_url(msg)
    if module_url is None:
        raise Exception('Unable to retrieve module URL from publish log.')

    module = get(module_url)

    if 'application/vnd.sas.microanalytic.module' in module._headers['content-type']:
        # Bind Python methods to the module instance that will execute the
        # corresponding MAS module step.
        from sasctl.services import microanalytic_score as mas

        return mas.define_steps(module)
    return module


def update_model_performance(data, model, label, refresh=True):
    """Upload data for calculating model performance metrics.

    Model performance and data distributions can be tracked over time by
    designating one or more tables that contain input data and target values.
    Performance metrics can be updated by uploading a data set for a new time
    period and executing the performance definition.

    Parameters
    ----------
    data : Dataframe
    model : str or dict
        The name or id of the model, or a dictionary representation of
        the model.
    label : str
        The time period the data is from.  Should be unique and will be
        displayed on performance charts.  Examples: 'Q1', '2019', 'APR2019'.
    refresh : bool, optional
        Whether to execute the performance definition and refresh results with
        the new data.

    Returns
    -------
    CASTable
        The CAS table containing the performance data.

    See Also
    --------
     :meth:`model_management.create_performance_definition <.ModelManagement.create_performance_definition>`

    .. versionadded:: v1.3

    """
    try:
        import swat
    except ImportError:
        raise RuntimeError(
            "The 'swat' package is required to save model " "performance data."
        )

    # Default to true
    refresh = True if refresh is None else refresh

    model_obj = mr.get_model(model)

    if model_obj is None:
        raise ValueError('Model %s was not found.' % model)

    project = mr.get_project(model_obj.projectId)

    if project.get('function', '').lower() not in ('prediction', 'classification'):
        raise ValueError(
            "Performance monitoring is currently supported for "
            "regression and binary classification projects.  "
            "Received project with '%s' function.  Should be "
            "'Prediction' or 'Classification'." % project.get('function')
        )

    if project.get('targetLevel', '').lower() not in ('interval', 'binary'):
        raise ValueError(
            "Performance monitoring is currently supported for "
            "regression and binary classification projects.  "
            "Received project with '%s' target level.  Should be "
            "'Interval' or 'Binary'." % project.get('targetLevel')
        )

    if (
        project.get('predictionVariable', '') == ''
        and project.get('function', '').lower() == 'prediction'
    ):
        raise ValueError(
            "Project '%s' does not have a prediction variable " "specified." % project
        )

    if (
        project.get('eventProbabilityVariable', '') == ''
        and project.get('function', '').lower() == 'classification'
    ):
        raise ValueError(
            "Project '%s' does not have an Event Probability variable "
            "specified." % project
        )

    # Find the performance definition for the model
    # As of Viya 3.4, no way to search by model or project
    perf_def = None
    for p in mm.list_performance_definitions():
        if project.id in p.projectId:
            perf_def = p
            break

    if perf_def is None:
        raise ValueError(
            "Unable to find a performance definition for model " "'%s'" % model
        )

    # Check where performance datasets should be uploaded
    cas_id = perf_def['casServerId']
    caslib = perf_def['dataLibrary']
    table_prefix = perf_def['dataPrefix']

    # All input variables must be present
    missing_cols = [col for col in perf_def.inputVariables if col not in data.columns]
    if missing_cols:
        raise ValueError(
            "The following columns were expected but not found in "
            "the data set: %s" % ', '.join(missing_cols)
        )

    # If CAS is not executing the model then the output variables must also be
    # provided
    if not perf_def.scoreExecutionRequired:
        missing_cols = [
            col for col in perf_def.outputVariables if col not in data.columns
        ]
        if missing_cols:
            raise ValueError(
                "The following columns were expected but not found in the data "
                "set: %s" % ', '.join(missing_cols)
            )

    sess = current_session()
    regex = r'{}_(\d+)_.*_{}'.format(table_prefix, model_obj.id)

    # Save the current setting before overwriting
    orig_sslreqcert = os.environ.get('SSLREQCERT')

    # If SSL connections to microservices are not being verified, don't attempt
    # to verify connections to CAS - most likely certs are not in place.
    if not sess.verify:
        os.environ['SSLREQCERT'] = 'no'

    # Upload the performance data to CAS
    with sess.as_swat(server=cas_id) as s:

        s.setsessopt(messagelevel='warning')

        with swat.options(exception_on_severity=2):
            caslib_info = s.table.tableinfo(caslib=caslib)

        all_tables = getattr(caslib_info, 'TableInfo', None)
        if all_tables is not None:
            # Find tables with similar names
            perf_tables = all_tables.Name.str.extract(
                regex, flags=re.IGNORECASE, expand=False
            )

            # Get last-used sequence number
            last_seq = perf_tables.dropna().astype(int).max()
            next_seq = 1 if math.isnan(last_seq) else last_seq + 1
        else:
            next_seq = 1

        table_name = '{prefix}_{sequence}_{label}_{model}'.format(
            prefix=table_prefix, sequence=next_seq, label=label, model=model_obj.id
        )

        with swat.options(exception_on_severity=2):
            # Table must be promoted so performance jobs can access.
            result = s.upload(
                data, casout=dict(name=table_name, caslib=caslib, promote=True)
            )

            if not hasattr(result, 'casTable'):
                raise RuntimeError('Unable to upload performance data to CAS.')

            tbl = result.casTable

    # Restore the original value
    if orig_sslreqcert is not None:
        os.environ['SSLREQCERT'] = orig_sslreqcert

    # Execute the definition if requested
    if refresh:
        mm.execute_performance_definition(perf_def)

    return tbl


def _parse_module_url(msg):
    try:
        details = json.loads(msg)

        module_url = get_link(details, 'module')
        module_url = module_url.get('href')
    except json.JSONDecodeError:
        match = re.search(r'(?:rel=module, href=(.*?),)', msg)  # Vya 3.5
        if match is None:
            match = re.search(
                r'(?:Rel: module URI: (.*?) MediaType)', msg
            )  # Format changed in Viya 4.0
        module_url = match.group(1) if match else None

    return module_url



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

if __name__ == '__main__':
    from sasctl.utils.cli import main

    main()




#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
import re
from collections import OrderedDict
from inspect import getfullargspec, getsourcelines

from .ds2 import DS2Variable
from ..decorators import versionchanged


logger = logging.getLogger(__name__)


@versionchanged('`names` parameter added.', version='1.5')
def ds2_variables(input, output_vars=False, names=None):
    """Generate a collection of `DS2Variable` instances corresponding to the input

    Parameters
    ----------
    input : function or OrderedDict<string, tuple> or Pandas DataFrame or Numpy or OrderedDict<string, type>
        a function or mapping parameter names to (type, is_output)

    output_vars : bool
        Whether or not to treat  all variables from `input` as output variables

    names : list of str
        List of variable names to used.  If a single string is specified it will
        be used as a prefix and variable names in the format "prefixNN" will be
        generated.

    Returns
    -------
    list<DS2Variable>

    Examples
    --------
    >>> ds2_variables(OrderedDict(a=int, c=float))
    [DS2Variable(name='a', type='integer', out=False), DS2Variable(name='c', type='double', out=False)]

    >>> ds2_variables({'x': (float, True)})
    [DS2Variable(name='x', type='double', out=True)]

    """

    if isinstance(input, dict):
        types = input
    elif hasattr(input, 'columns') and hasattr(input, 'dtypes'):
        # Pandas DataFrame
        types = OrderedDict()
        for col in input.columns:
            if input[col].dtype.name == 'object':
                types[col] = ('char', False)
            elif input[col].dtype.name == 'category':
                types[col] = ('char', False)
            else:
                types[col] = (input[col].dtype.name, False)
    elif hasattr(input, 'dtype'):
        # Numpy array?  No column names, but we can at least create dummy vars of the correct type
        types = OrderedDict(
            [
                ('var{}'.format(i), (input.dtype.name.replace('object', 'char'), False))
                for i in range(1, input.size + 1)
            ]
        )
    elif callable(input):
        types = parse_type_hints(input)
    else:
        raise RuntimeError(
            "Unable to determine input/ouput types using "
            "instance of type '%s'." % type(input)
        )

    if isinstance(names, str):
        names = [names + str(i) for i in range(1, len(types) + 1)]
    elif names is None:
        names = list(types.keys())

    results = []
    for v in types.values():
        name = names.pop(0)
        if isinstance(v, str):
            results.append(DS2Variable(name=name, type=v, out=output_vars))
        elif isinstance(v, type):
            results.append(DS2Variable(name=name, type=v.__name__, out=output_vars))
        elif isinstance(v, tuple):
            type_ = v[0].__name__ if isinstance(v[0], type) else str(v[0])
            out = v[1] or output_vars
            results.append(DS2Variable(name=name, type=type_, out=out))
        else:
            raise RuntimeError('Unable to determine input/ouput types.')

    return results


def parse_type_hints(func, skip_var='self'):
    """Attempt to discern types for the input and output variable(s).

    DS2 is a strongly-typed language but Python is not.  Need to determine the types for input/output variables so they
    can be correctly mapped to DS2 types.

    Parameters
    ----------
    func : function
        the object to inspect for parameters

    skip_var : str
        name of the variable assumed to be the instance reference, if any.  Will be ignored during parameter parsing.

    Returns
    -------
    params : OrderedDict
        dictionary of str : (str, bool) mapping param_name : (param_type, is_return_val)

    Raises
    ------
    ValueError
        If the type of one or more variables could not be determined.

    """

    params = OrderedDict(
        [(k, None) for k in getfullargspec(func).args if k != skip_var]
    )
    logger.debug('Params: {}'.format(params))

    if getattr(func, '__annotations__', None):
        params.update(parse_type_hints_from_annotations(func, skip_var=skip_var))
    else:
        params.update(parse_type_hints_from_source(func, skip_var=skip_var))

    if any(v is None for v in params.values()):
        raise ValueError('Unable to determine parameter types.')

    return params


def parse_type_hints_from_source(func, skip_var='self'):
    """Parse type hints stored in comments according to PEP 484."""

    regex = re.compile(r'^\s+\# types?: ', re.IGNORECASE)

    def parse_types(line):
        if line:
            line = regex.sub('', line)  # Strip out the ' # type:' portion if it exists
            return line.strip().strip('(').strip(')').split(',')
        return []

    params = OrderedDict()

    for line in getsourcelines(func)[0]:
        if regex.match(line):
            if '->' in line:
                inputs, outputs = line.split('->')
            else:
                inputs = line
                outputs = None

            types = parse_types(inputs)
            args = getfullargspec(func).args  # skipcq PYL-W1505

            try:
                args.remove(skip_var)
            except ValueError:
                pass

            for a, t in zip(args, types):
                params[a] = (t.strip(), False)

            types = parse_types(outputs)
            for i, t in enumerate(types):
                params['out%d' % (i + 1)] = (t.strip(), True)

    return params


def parse_type_hints_from_annotations(func, skip_var='self'):
    """Parse type hints from the function signature."""

    annotations = func.__annotations__
    params = OrderedDict()
    logger.debug('Annotations: {}'.format(annotations))

    for p, t in annotations.items():
        if p == skip_var:
            continue
        elif p == 'return':
            if t is not None:
                params[p] = (t.__name__, True)
        else:
            params[p] = (t.__name__, False)

    return params



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service
from ..utils.decorators import experimental

class ModelManagement(Service):
    """The Model Management API provides basic resources for monitoring
    performance, comparing models, and running workflow processes.
    """
    _SERVICE_ROOT= '/modelManagement'


    list_performance_definitions, get_performance_definition, \
        update_performance_definition, delete_performance_definition = \
        Service._crud_funcs('/performanceTasks', 'performance task',
                            'performance tasks')

    # TODO:  set ds2MultiType
    @classmethod
    def publish_model(cls,
                      model,
                      destination,
                      name=None,
                      force=False,
                      reload_model_table=False):
        """

        Parameters
        ----------
        model
        destination
        name
        force : bool, optional
            Whether to overwrite the model if it already exists in the
            publishing `destination`.
        reload_model_table : bool, optional
            Whether the model table in CAS should be reloaded.  Defaults to
            False.

        Returns
        -------

        """
        from .model_repository import ModelRepository
        from .model_publish import ModelPublish

        mr = ModelRepository()
        mp = ModelPublish()

        model_obj = mr.get_model(model)

        if model_obj is None:
            model_name = model.name if hasattr(model, 'name') else str(model)
            raise ValueError("Model '{}' was not found.".format(model_name))

        model_uri = mr.get_model_link(model_obj, 'self')

        # TODO: Verify allowed formats by destination type.
        # As of 19w04 MAS throws HTTP 500 if name is in invalid format.
        model_name = name or '{}_{}'.format(model_obj['name'].replace(' ', ''),
                                            model_obj['id']).replace('-', '')

        request = {
            "name": model_obj.get('name'),
            "notes": model_obj.get('description'),
            "modelContents": [
                {
                    "modelName": mp._publish_name(
                        model_obj.get('name')),
                    "sourceUri": model_uri.get('uri'),
                    "publishLevel": "model"
                }
            ],
            "destinationName": destination
        }

        # Publishes a model that has already been registered in the model
        # repository.
        # Unlike model_publish service, does not require Code to be specified.
        r = cls.post('/publish',
                     json=request,
                     params=dict(force=force,
                                  reloadModelTable=reload_model_table),
                     headers={'Content-Type':
                                   'application/vnd.sas.models.publishing.request.asynchronous+json'})
        return r

    @classmethod
    def create_performance_definition(cls,
                                      model,
                                      library_name,
                                      table_prefix,
                                      name=None,
                                      description=None,
                                      monitor_champion=False,
                                      monitor_challenger=False,
                                      max_bins=None,
                                      scoring_required=False,
                                      all_data=False,
                                      save_output=True,
                                      output_library=None,
                                      autoload_output=False,
                                      cas_server=None,
                                      trace=False):
        """Create the performance task definition in the model project to
        monitor model performance.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of the model.
        library_name : str
            The library containing the input data, default is 'Public'.
        table_prefix : str
            The name used for the performance data.
        name : str
            The name of the performance task.
        description : str
            The description of the performance task, default is 'Performance
            monitoring for model' + model.name.
        monitor_champion : bool
            Indicates to monitor the project champion model.
        monitor_challenger : bool
            Indicates to monitor challenger models.
        max_bins : int
            The maximum bins number, Must be >= 2.  Defaults to 10.
        scoring_required : bool
            Whether model scoring must be performed on the input data before
            performance results can be computed.  Should be `False` if target
            values are included in the `table_prefix` tables.
        all_data : bool
            Whether to run the performance job against all matching data tables
            in `library_name` or just the new tables.  Defaults to `False`.
        save_output : bool
            Whether to save the computed results to a table in `output_library`.
            Defaults to True.
        output_library : str
            Name of a CASLIB where computed results should be saved.  Defaults to
            'ModelPerformanceData'.
        autoload_output : bool
            Whether computed results should automatically be re-loaded
            after a CAS server restart.
        cas_server : str
            The CAS Server for the monitoring task, default is 'cas-shared-default'.
        trace : bool
            Whether to enable trace messages in the SAS job log when
            executing the performance definition.

        Returns
        -------
        RestObj
            The performance task definition schema

        """
        from .model_repository import ModelRepository

        if not scoring_required and '_' in table_prefix:
            raise ValueError(
                "Parameter 'table_prefix' cannot contain underscores."
                " Received a value of '%s'.") % table_prefix

        max_bins = 10 if max_bins is None else int(max_bins)
        if int(max_bins) < 2:
            raise ValueError("Parameter 'max_bins' must be at least 2.  "
                             "Received a value of '%s'." % max_bins)

        mr = ModelRepository()
        model = mr.get_model(model)
        project = mr.get_project(model.projectId)

        # Performance data cannot be captured unless certain project properties
        # have been configured.
        for required in ['targetVariable', 'targetLevel']:
            if getattr(project, required, None) is None:
                raise ValueError("Project %s must have the '%s' property set."
                                 % (project.name, required))
        if project.get('function') == 'classification' \
                and project.get('eventProbabilityVariable') is None:
            raise ValueError("Project %s must have the "
                             "'eventProbabilityVariable' property set."
                             % project.name)
        if project.get('function') == 'prediction' \
                and project.get('predictionVariable') is None:
            raise ValueError("Project %s must have the 'predictionVariable' "
                             "property set." % project.name)

        request = {'projectId': project.id,
                   'name': name or model.name + ' Performance',
                   'modelIds': [model.id],
                   'championMonitored': monitor_champion,
                   'challengerMonitored': monitor_challenger,
                   'maxBins': max_bins,
                   'resultLibrary': output_library or 'ModelPerformanceData',
                   'includeAllData': all_data,
                   'scoreExecutionRequired': scoring_required,
                   'performanceResultSaved': save_output,
                   'loadPerformanceResult': autoload_output,
                   'dataLibrary': library_name or 'Public',
                   'description': description or 'Performance definition for model ' + model.name,
                   'casServerId': cas_server or 'cas-shared-default',
                   'dataPrefix': table_prefix,
                   'traceOn': trace
                   }

        # If model doesn't specify input/output variables, try to pull from project definition
        if model.get('inputVariables', []):
            request['inputVariables'] = [v.get('name') for v in
                                         model['inputVariables']]
            request['outputVariables'] = [v.get('name') for v in
                                          model['outputVariables']]
        else:
            request['inputVariables'] = [v.get('name') for v in
                                         project.get('variables', []) if
                                         v.get('role') == 'input']
            request['outputVariables'] = [v.get('name') for v in
                                          project.get('variables', []) if
                                          v.get('role') == 'output']

        return cls.post('/performanceTasks', json=request,
                        headers={
            'Content-Type': 'application/vnd.sas.models.performance.task+json'})

    @classmethod
    def execute_performance_definition(cls, definition):
        """Launches a job to run a performance definition.

        Parameters
        ----------
        definition : str or dict
            The id or dictionary representation of a performance definition.

        Returns
        -------
        RestObj
            The executing job

        """
        definition = cls.get_performance_definition(definition)

        return cls.post('/performanceTasks/%s' % definition.id)

    @classmethod
    @experimental
    def list_model_workflow_definition(cls):
        """List all enabled Workflow Processes to execute on Model Project.

        Returns
        -------
        RestObj
            The list of workflows

        """
        from .workflow import Workflow
        wf = Workflow()

        return wf.list_enabled_definitions()

    @classmethod
    @experimental
    def list_model_workflow_prompt(cls, workflowName):
        """List prompt Workflow Processes Definitions.

        Parameters
        ----------
        workflowName : str
            Name or ID of an enabled workflow to retrieve inputs

        Returns
        -------
        list
            The list of prompts for specific workflow
    
        """
        from .workflow import Workflow
        wf = Workflow()
        
        return wf.list_workflow_prompt(workflowName)

    @classmethod
    @experimental
    def list_model_workflow_executed(cls, projectName):
        """List prompt Workflow Processes Definitions.

        Parameters
        ----------
        projectName : str
            Name of the Project list executed workflow

        Returns
        -------
        RestObj
            List of workflows associated to project

        """
        from .model_repository import ModelRepository
        mr = ModelRepository()

        project = mr.get_project(projectName)

        return cls.get('/workflowProcesses?filter=eq(associations.solutionObjectId,%22' + project['id'] + '%22)')

    @classmethod
    @experimental
    def execute_model_workflow_definition(cls, project_name, workflow_name, input=None):
        """Runs specific Workflow Processes Definitions.

        Parameters
        ----------
        project_name : str
            Name of the Project that will execute workflow
        workflow_name : str
            Name or ID of an enabled workflow to execute
        input : dict, optional
            Input values for the workflow for initial workflow prompt

        Returns
        -------
        RestObj
            The executing workflow
            
        """
        from .model_repository import ModelRepository
        from .workflow import Workflow

        mr = ModelRepository()
        wf = Workflow()
        
        project = mr.get_project(project_name)

        workflow = wf.run_workflow_definition(workflow_name, input=input)
        
        # Associations running workflow to model project, note workflow has to be running
        # THINK ABOUT: do we do a check on status of the workflow to determine if it is still running before associating?

        input = {"processName": workflow['name'],
                 "processId": workflow['id'],
                 "objectType": "MM_Project",
                 "solutionObjectName": project_name,
                 "solutionObjectId": project['id'],
                 "solutionObjectUri": "/modelRepository/projects/" + project['id'],
                 "solutionObjectMediaType": "application/vnd.sas.models.project+json"}
        
        #Note, you can get a HTTP Error 404: {"errorCode":74052,"message":"The workflow process for id <> cannot be found.
        #                                    Associations can only be made to running processes.","details":["correlator:
        #                                    e62c5562-2b11-45db-bcb7-933200cb0f0a","traceId: 3118c0fb1eb9702d","path:
        #                                    /modelManagement/workflowAssociations"],"links":[],"version":2,"httpStatusCode":404} 
        # Which is fine and expected like the Visual Experience.
        return cls.post('/workflowAssociations',
                        json=input,
                        headers={'Content-Type': 'application/vnd.sas.workflow.object.association+json'})





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""A stateless, memory-resident, high-performance program execution service."""

import re
from collections import OrderedDict
from math import isnan

from .service import Service


class MicroAnalyticScore(Service):
    """Micro Analytic Service (MAS) client."""

    _SERVICE_ROOT = '/microanalyticScore'

    @classmethod
    def is_uuid(cls, id_):
        """Check if the ID appears to be a valid MAS id.

        Indicates whether `id` appears to be a correctly formatted ID.  Does
        **not** check whether a module with `id` actually exists.

        Parameters
        ----------
        id_ : str

        Returns
        -------
        bool

        Notes
        -----
        Overrides the :meth:`Service.is_uuid` method since MAS modules do
        not currently use IDs that are actually UUIDs.

        """
        # Anything that consists of only numbers, lowercase letters,
        # and underscores, and does not start with a number, looks like a
        # MAS id.
        return re.match('^[_a-z][_a-z0-9]+$', id_) is not None

    list_modules, get_module, update_module, delete_module = Service._crud_funcs(
        '/modules', 'module'
    )

    def get_module_step(self, module, step):
        """Details of a single step in a given module.

        Parameters
        ----------
        module : str or dict
            Name, id, or dictionary representation of a module
        step : str
            Name of the step

        Returns
        -------
        RestObj

        """
        module = self.get_module(module)

        r = self.get('/modules/{}/steps/{}'.format(module.id, step))
        return r

    def list_module_steps(self, module):
        """List all steps defined for a module.

        Parameters
        ----------
        module : str or dict
            Name, id, or dictionary representation of a module

        Returns
        -------
        list
            List of :class:`.RestObj` instances representing each step.

        """
        module = self.get_module(module)

        steps = self.get('/modules/{}/steps'.format(module.id))
        return steps if isinstance(steps, list) else [steps]

    def execute_module_step(self, module, step, return_dict=True, **kwargs):
        """Call a module step with the given parameters.

        Parameters
        ----------
        module : str or dict
            Name, id, or dictionary representation of a module
        step : str
            Name of the step
        return_dict : bool, optional
            Whether the results should be returned as a dictionary instead
            of a tuple
        kwargs : any
            Passed as arguments to the module step

        Returns
        -------
        any
            Results of the step execution.  Returned as a dictionary if
            `return_dict` is True, otherwise returned as a tuple if more
            than one value is returned, otherwise the single value.

        """
        module_name = module.name if hasattr(module, 'name') else str(module)
        module = self.get_module(module)

        if module is None:
            raise ValueError("Module '{}' was not found.".format(module_name))
        module = module.id
        step = step.id if hasattr(step, 'id') else step

        # Make sure all inputs are JSON serializable
        # Common types such as numpy.int64 and numpy.float64 are NOT
        # serializable
        for k in kwargs:
            type_name = type(kwargs[k]).__name__
            if type_name == 'float64':
                kwargs[k] = float(kwargs[k])
            elif type_name == 'int64':
                kwargs[k] = int(kwargs[k])

        body = {'inputs': [{'name': k, 'value': v} for k, v in kwargs.items()]}

        # Convert NaN to None (null) before calling MAS
        for i in body['inputs']:
            try:
                if isnan(i['value']):
                    i['value'] = None
            except TypeError:
                pass

        r = self.post('/modules/{}/steps/{}'.format(module, step), json=body)

        # Convert list of name/value pair dictionaries to single dict
        outputs = OrderedDict()
        for output in r.get('outputs', []):
            k, v = output['name'], output.get('value')

            # Remove padding from CHAR columns
            if isinstance(v, str):
                v = v.strip()

            outputs[k] = v

        if return_dict:
            # Return results as k=v pairs
            return outputs

        # Return only the values, as if calling another Python function.
        outputs = tuple(outputs.values())
        if len(outputs) == 1:
            return outputs[0]
        return outputs

    def create_module(
        self,
        name=None,
        description=None,
        source=None,
        language='python',
        scope='public',
    ):
        """Create a new module in MAS.

        Parameters
        ----------
        name : str
        description : str
        source : str
        language : str { 'python', 'ds2' }
        scope : str { 'public', 'private' }

        Returns
        -------
        RestObj

        """
        if source is None:
            raise ValueError('The `source` parameter is required.')
        source = str(source)

        if language == 'python':
            t = 'text/x-python'
        elif language == 'ds2':
            t = 'text/vnd.sas.source.ds2'
        else:
            raise ValueError('Unrecognized source code language `%s`.' % language)

        data = {
            'id': name,
            'type': t,
            'description': description,
            'source': source,
            'scope': scope,
        }

        r = self.post('/modules', json=data)
        return r

    def define_steps(self, module):
        """Map MAS steps to Python methods.

        Defines python methods on a module that automatically call the
        corresponding MAS steps.

        Parameters
        ----------
        module : str or dict
            Name, id, or dictionary representation of a module

        Returns
        -------
        RestObj
            The module with additional methods defined.

        """
        import types

        module = self.get_module(module)

        # Define a method for each step of the module
        for id_ in module.get('stepIds', []):
            step = self.get_module_step(module, id_)

            # Method should have an argument for each parameter of the step
            arguments = [k['name'] for k in step.get('inputs', [])]
            arg_types = [k['type'] for k in step.get('inputs', [])]

            # Format call to execute_module_step()
            call_params = ['{arg}={arg}'.format(arg=a) for a in arguments]

            # Set type hints for the function
            type_string = '    # type: ({})'.format(', '.join(arg_types))

            # Method signature
            # Default all params to None to allow method(DataFrame) execution
            input_params = [a for a in arguments] + ['**kwargs']
            method_name = '_%s_%s' % (module.id, step.id)
            signature = 'def %s(%s):' % (method_name, ', '.join(input_params))

            # If the module step takes arguments, then we perform a number
            # of additional checks to make the allowed input as wide as possible
            if arguments:
                # Perform some initial checks to see if the input is a Numpy array
                # or a Pandas DataFrame.  Each will be handled separately.
                arg_checks = [
                    'first_input = %s' % arguments[0],
                    'first_input_type = str(type(first_input))',
                    'try:',
                    '    import pandas as pd',
                    '    if isinstance(first_input, pd.DataFrame):',
                    '        assert first_input.shape[0] == 1',
                    '        first_input = first_input.iloc[0]',
                    '    is_pandas = isinstance(first_input, pd.Series)',
                    'except ImportError:',
                    '    is_pandas = False',
                    'try:',
                    '    import numpy as np',
                    '    is_numpy = isinstance(first_input, np.ndarray)',
                    'except ImportError:',
                    '    is_numpy = False',
                    'if is_pandas:',
                    '    assert len(first_input.shape) == 1',
                    '    for k in first_input.keys():',
                ]

                for arg in arguments:
                    arg_checks.append("        if k.lower() == '%s':" % arg.lower())
                    arg_checks.append("            %s = first_input[k]" % arg)
                    arg_checks.append("            continue")

                arg_checks.extend(
                    [
                        'elif is_numpy:',
                        '    if len(first_input.shape) > 1:',
                        '        assert first_input.shape[0] == 1',
                        '    first_input = first_input.ravel()',
                    ]
                )

                for i, arg in enumerate(arguments):
                    arg_checks.append('    %s = first_input[%d]' % (arg, i))

                # MAS always lower-cases variable names
                # Since the original Python variables may have a different case,
                # allow kwargs to be used to input alternative caps
                arg_checks.append('else:')
                arg_checks.append('    for k in kwargs.keys():')
                for arg in arguments:
                    arg_checks.append("        if k.lower() == '%s':" % arg.lower())
                    arg_checks.append("            %s = kwargs[k]" % arg)
                    arg_checks.append("            continue")

                # MAS does not attempt any type conversions, so int, decimal, etc
                # needs to be exactly typed or the call will fail.  Cast everything
                # just to be sure.
                for arg, arg_type in zip(arguments, arg_types):
                    if arg_type == 'decimal':
                        arg_checks.append('%s = float(%s)' % (arg, arg))
                    elif arg_type == 'integer':
                        arg_checks.append('%s = int(%s)' % (arg, arg))
            else:
                arg_checks = []

            # Full method source code
            # Drops 'rc' and 'msg' from return values
            code = (
                signature,
                type_string,
                '    """Execute step \'%s\' of module \'%s\'."""' % (step, module),
                '\n'.join('    %s' % a for a in arg_checks),
                '    r = execute_module_step(%s)'
                % ', '.join(['module', 'step'] + call_params),
                '    r.pop("rc", None)',
                '    r.pop("msg", None)',
                '    if len(r) == 1:',
                '        return r.popitem()[1]',
                '    return tuple(v for v in r.values())',
            )

            code = '\n'.join(code)
            self.log.debug(
                "Generated code for step '%s' of module '%s':\n%s", id_, module, code
            )
            compiled = compile(code, '<string>', 'exec')

            env = globals().copy()
            env.update(
                {
                    'execute_module_step': self.execute_module_step,
                    'module': module,
                    'step': step,
                }
            )

            func = types.FunctionType(
                compiled.co_consts[0], env, argdefs=tuple(None for x in arguments)
            )

            setattr(module, step.id, func)

        return module



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service


class Relationships(Service):
    """The Relationships API manages a repository of relationships.

    A relationship describes how two resources are connected. A relationship
    contains a subject resource, the related resource, and the relationship
    type.  A relationship type describes the nature of the relationship between
    two resources. The relationship type includes a name, label, and
    description.

    """

    _SERVICE_ROOT = '/relationships'

    list_relationships, get_relationship, update_relationship, \
        delete_relationship = Service._crud_funcs('/relationships', 'relationship')



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""The Model Repository service supports registering and managing models."""

from warnings import warn

from .service import Service
from ..core import current_session, get, delete, sasctl_command, HTTPError

FUNCTIONS = {
    'Analytical',
    'Classification',
    'Clustering',
    'Forecasting',
    'Prediction',
    'Text categorization',
    'Text extraction',
    'Text sentiment',
    'Text topics',
    'Sentiment',
}


def _get_filter(x):
    # Model Repository filtering is done using the properties= query parameter
    # instead of the default filter= parameter (as of Viya 3.4).
    # Define a custom function for building out the filter
    return dict(properties='(name, %s)' % x)


class ModelRepository(Service):
    """Implements the Model Repository REST API.

    The Model Repository API provides support for registering, organizing,
    and managing models within a common model repository.

    See Also
    --------
    `REST Documentation <https://developer.sas.com/apis/rest/DecisionManagement
    /#model-repository>`_

    """

    _SERVICE_ROOT = '/modelRepository'

    list_repositories, _, update_repository, delete_repository = Service._crud_funcs(
        '/repositories', 'repository', get_filter=_get_filter
    )

    list_projects, get_project, update_project, delete_project = Service._crud_funcs(
        '/projects', 'project', get_filter=_get_filter
    )

    list_models, get_model, update_model, delete_model = Service._crud_funcs(
        '/models', 'model', get_filter=_get_filter
    )

    @classmethod
    def get_astore(cls, model):
        """Get the ASTORE for a model registered int he model repository.

        Parameters
        ----------
        model :  str or dict
            The name or id of the model, or a dictionary representation of the
            model.

        Returns
        ----------
        binary?

        """
        # TODO: Download binary object?

        link = cls.get_model_link(model, 'analyticStore', refresh=True)

        if link is not None:
            return link.get('href')

    @classmethod
    def get_model_link(cls, model, rel, refresh=False):
        """Retrieve a link from a model's set of links.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of the
            model.
        rel : str
            The name of the link to retrieve
        refresh : bool, optional
            Whether the model's data should be refreshed before searching for
            the link.

        Returns
        -------
        dict or None
            Dictionary containing the link's properties

        """
        # Try to find the link if the model is a dictionary
        if isinstance(model, dict):
            link = cls.get_link(model, rel)
            if link is not None or not refresh:
                return link

        # Model either wasn't a dictionary or it didn't include the link
        # Pull the model data and then check for the link
        model = cls.get_model(model, refresh=refresh)

        return cls.get_link(model, rel)

    @classmethod
    def get_score_code(cls, model):
        """Get the score code for a model registered in the model repository.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model.

        Returns
        -------
        str
            The code designated as the model's score code

        """
        link = cls.get_model_link(model, 'scoreCode', refresh=True)

        if link is not None:
            scorecode_uri = link.get('href')
            return get(
                scorecode_uri,
                headers={'Accept': 'text/vnd.sas.models.score.code.ds2package'},
            )

    @classmethod
    def get_model_contents(cls, model):
        """Retrieve the additional files and data associated with the model.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of the
            model.

        Returns
        -------
        list
            The code designated as the model's score code

        """

        link = cls.get_model_link(model, 'contents', refresh=True)
        contents = cls.request_link(link, 'contents')

        # By default, request_link() will unwrap a length-1 list.
        # If that happens, re-wrap so a list is always returned.
        if isinstance(contents, list):
            return contents

        return [contents]

    @classmethod
    @sasctl_command('get', 'repositories')
    def get_repository(cls, repository, refresh=False):
        """Return a repository instance.

        Parameters
        ----------
        repository : str or dict
            Name, ID, or dictionary representation of the repository.
        refresh : bool, optional
            Obtain an updated copy of the repository.

        Returns
        -------
        RestObj or None
            A dictionary containing the repository attributes or None.

        Notes
        -------
        If `repository` is a complete representation of the repository it will be
        returned unless `refresh` is set.  This prevents unnecessary REST
        calls when data is already available on the client.

        """
        # If the input already appears to be the requested object just
        # return it, unless a refresh of the data was explicitly requested.
        if isinstance(repository, dict) and all(
            k in repository for k in ('id', 'name')
        ):
            if refresh:
                repository = repository['id']
            else:
                return repository

        if cls.is_uuid(repository):
            try:
                # Attempt to GET the repository directly.  Access may be restricted, so allow HTTP 403 errors
                # and fall back to using list_repositories() instead.
                return cls.get('/repositories/{id}'.format(id=repository))
            except HTTPError as e:
                if e.code != 403:
                    raise e

        results = cls.list_repositories()

        # Not sure why, but as of 19w04 the filter doesn't seem to work
        for result in results:
            if result['name'] == str(repository) or result['id'] == str(repository):
                # Make a request for the specific object so that ETag
                # is included, allowing updates.
                try:
                    if cls.get_link(result, 'self'):
                        return cls.request_link(result, 'self')

                    id_ = result.get('id', result['name'])
                    return cls.get('/repositories/{id}'.format(id=id_))
                except HTTPError as e:
                    # NOTE: As of Viya 4.0.1 access to GET a repository is restricted to admin users out of the box.
                    # Try to GET the repository, but ignore any 403 (permission denied) errors.
                    if e.code != 403:
                        raise e
                return result

    @classmethod
    def create_model(
        cls,
        model,
        project,
        description=None,
        modeler=None,
        function=None,
        algorithm=None,
        tool=None,
        score_code_type=None,
        training_table=None,
        event_prob_variable=None,
        event_target_value=None,
        is_champion=False,
        is_challenger=False,
        location=None,
        target_variable=None,
        is_retrainable=False,
        is_immutable=False,
        properties=None,
        input_variables=None,
        output_variables=None,
    ):
        """Create a model in an existing project or folder.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of the
            model.
        project : str or dict
            The name or id of the model project, or a dictionary representation
            of the model project.
        description : str, optional
            The description of the model.
        modeler : str, optional
            Name of the user that created the model.  Current user name will be
            used if unspecified.
        function : str, optional
            The function of the model, valid values include: analytical,
            classification, cluster, forecasting, prediction, Text analytics,
            transformation.
        algorithm : str, optional
            The name of the model algorithm.
        tool : str, optional
            The name of the model tool, can be 'Python 2' or 'Python 3'.
        score_code_type : str, optional
            The score code type for the model.
        training_table : str, optional
            The train data table.
        event_prob_variable : str, optional
            The name of the event probability variable.  Used for
            classification models only.
        event_target_value : str, optional
            The target event value.  Used for classification models only.
        is_champion : bool, optional
            Indicates whether the model should be designated as the
            project's champion model.  Defaults to False.
        is_challenger : bool, optional
            Indicates whether the model should be designated as a
            challenger model in the project.  Defaults to False.
        location : str, optional,
            The location of this model.
        target_variable : str, optional
            The name of the target variable.
        is_retrainable : bool
            Indicates whether the model can be retrained or not.
        is_immutable : bool
            Indicates whether the model can be changed or not.
        properties : array_like, optional (custom properties)
            Custom model properties that can be set: name, value, type
        input_variables : array_like, optional
            Model input variables. By default, these are the same as the model
            project.
        output_variables : array_like, optional
            Model output variables. By default, these are the same as the model
             project.

        Returns
        -------
        str
            The model schema returned in JSON format.

        """
        properties = properties or {}

        if isinstance(model, str):
            model = {'name': model}

        p = cls.get_project(project)
        if p is None:
            raise ValueError("Unable to find project '%s'" % project)

        # Use any explicitly passed parameter value first.
        # Fall back to values in the model dict.
        model['projectId'] = p['id']
        model['modeler'] = modeler or model.get('modeler') or current_session().username
        model['description'] = description or model.get('description')
        model['function'] = function or model.get('function')
        model['algorithm'] = algorithm or model.get('algorithm')
        model['tool'] = tool or model.get('tool')
        model['champion'] = is_champion or model.get('champion')

        if is_champion:
            model['role'] = 'champion'
        elif is_challenger:
            model['role'] = 'challenger'

        model.setdefault(
            'properties', [{'name': k, 'value': v} for k, v in properties.items()]
        )
        model['scoreCodeType'] = score_code_type or model.get('scoreCodeType')
        model['trainTable'] = training_table or model.get('trainTable')
        model[
            'classificationEventProbabilityVariableName'
        ] = event_prob_variable or model.get(
            'classificationEventProbabilityVariableName'
        )
        model['classificationTargetEventValue'] = event_target_value or model.get(
            'classificationTargetEventValue'
        )
        model['location'] = location or model.get('location')
        model['targetVariable'] = target_variable or model.get('targetVariable')
        model['retrainable'] = is_retrainable or model.get('retrainable')
        model['immutable'] = is_immutable or model.get('immutable')
        model['inputVariables'] = input_variables or model.get('inputVariables', [])
        model['outputVariables'] = output_variables or model.get('outputVariables', [])
        model['version'] = '2'

        return cls.post(
            '/models',
            json=model,
            headers={'Content-Type': 'application/vnd.sas.models.model+json'},
        )

    @classmethod
    def add_model_content(cls, model, file, name, role=None, content_type=None):
        """Add additional files to the model.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model.
        file : str or bytes
            A file related to the model, such as the model code.
        name : str
            Name of the file related to the model.
        role : str
            Role of the model file, such as 'Python pickle'.
        content_type : str
            an HTTP Content-Type value

        Returns
        -------
        str
            The model content schema.

        """
        if cls.is_uuid(model):
            id_ = model
        elif isinstance(model, dict) and 'id' in model:
            id_ = model['id']
        else:
            model = cls.get_model(model)
            id_ = model['id']

        if content_type is None and isinstance(file, bytes):
            content_type = 'application/octet-stream'

        if content_type is not None:
            files = {name: (name, file, content_type)}
        else:
            files = {name: file}

        metadata = {'role': role, 'name': name}

        # return cls.post('/models/{}/contents'.format(id_), files=files, data=metadata)

        # if the file already exists, a 409 error will be returned
        try:
            return cls.post(
                '/models/{}/contents'.format(id_), files=files, data=metadata
            )
        # delete the older duplicate model and rerun the API call
        except HTTPError as e:
            if e.code == 409:
                model_contents = cls.get_model_contents(id_)
                for item in model_contents:
                    if item.name == name:
                        cls.delete('/models/{}/contents/{}'.format(id_, item.id))
                        return cls.post(
                            '/models/{}/contents'.format(id_),
                            files=files,
                            data=metadata,
                        )
            else:
                raise e

    @classmethod
    def default_repository(cls):
        """Get the built-in default repository.

        Returns
        -------
        RestObj

        """
        all_repos = cls.list_repositories()

        if all_repos:
            # If nothing else, return the first repository
            repo = all_repos[0]

            # Check repository names to find a better default.
            # 'Repository 1' was default in 19w04
            # 'Public' was default in 19w21
            for r in all_repos:
                if r.name in ('Repository 1', 'Public'):
                    repo = r
                    break
            return repo

    @classmethod
    def create_project(cls, project, repository, **kwargs):
        """Create a model project in the given model repository.

        Parameters
        ----------
        project : str or dict
            The name or id of the model project, or a dictionary representation
            of the project.
        repository : str or dict
            The name or id of the model repository, or a dictionary
            representation of the repository.

        Returns
        -------
        RestObj

        """
        if isinstance(project, str):
            project = {'name': project}

        repository = cls.get_repository(repository)

        project['repositoryId'] = repository['id']
        project['folderId'] = repository['folderId']

        project.update(kwargs)
        return cls.post(
            '/projects',
            json=project,
            headers={'Content-Type': 'application/vnd.sas.models.project+json'},
        )

    @classmethod
    def import_model_from_zip(
        cls, name, project, file, description=None, version='latest'
    ):
        """Import a model and contents as a ZIP file into a model project.

        Parameters
        ----------
        name : str or dict
            The name of the model.
        project : str or dict
            The name or id of the model project, or a dictionary
            representation of the project.
        file : bytes
            The ZIP file containing the model and contents.
        description : str
            The description of the model.

        Returns
        -------
        RestObj
            The API response after importing the model.

        """
        project = cls.get_project(project)

        params = {
            'name': name,
            'description': description,
            'type': 'ZIP',
            'projectId': project.id,
            'versionOption': version,
        }
        params = '&'.join('{}={}'.format(k, v) for k, v in params.items())

        r = cls.post(
            '/models#octetStream',
            data=file.read(),
            params=params,
            headers={'Content-Type': 'application/octet-stream'},
        )
        return r

    @classmethod
    def create_model_version(cls, model, minor=False):
        """Create a new version of an existing model.

        Create a new major (X.0) or minor (1.X) version of an existing
        model.  All contents from the current version are copied to the new
        version.

        Parameters
        ----------
        model : str or dict
            The name, id, or dictionary representation of a model.
        minor : bool, optional
            Whether the new version should be a minor version increment from
            the current version.  Otherwise, a new major version will be
            created.  Defaults to False.

        Returns
        -------
        RestObj
            The new version of the model.

        """
        model_obj = cls.get_model(model)
        if cls.get_model_link(model_obj, 'addModelVersion') is None:
            raise ValueError("Unable to create a new version for model '%s'" % model)

        option = 'minor' if minor else 'major'

        # As of 8/9/19 addModelVersion returns the CURRENT model version,
        # not the NEW version.  Refresh the model to get the latest version.
        cls.request_link(model_obj, 'addModelVersion', json={'option': option})

        return cls.get_model(model_obj, refresh=True)

    @classmethod
    def list_model_versions(cls, model):
        """Get a list of previous versions of a model.

        The current model version is not included in the results.

        Parameters
        ----------
        model : str or dict
            The name, id, or dictionary representation of a model.

        Returns
        -------
        list

        """
        model = cls.get_model(model)
        if cls.get_model_link(model, 'modelVersions') is None:
            raise ValueError("Unable to retrieve versions for model '%s'" % model)

        return cls.request_link(model, 'modelVersions')

    @classmethod
    def copy_analytic_store(cls, model):
        """Copy model ASTOREs to a pre-defined server location.

        Copies all of the analytic stores for a model to the pre-defined
        server location (/config/data/modelsvr/astore).
        To enable publishing and scoring, models that contain analytic stores
        need the ASTORE files to be copied to a set location
        (/config/data/modelsrv/astore).  This location is used for
        integration with Event Stream Processing and others. This request
        invokes an asynchronous call to copy the analytic store files. Check
        the individual analytic store uris to get the completion state:
        pending, copying, success, failure. Please review the full Model
        Manager documentation before using.

        Parameters
        ----------
        model : str or dict
            The name, id, or dictionary representation of a model.

        Returns
        -------
        RestObj

        """
        rel = 'copyAnalyticStore'

        model = cls.get_model(model)

        if cls.get_link(model, rel) is None:
            model = cls.get_model(model, refresh=True)

        return cls.request_link(model, 'copyAnalyticStore')

    @classmethod
    def delete_model_contents(cls, model):
        """Deletes all contents (files) in the model.

        Parameters
        ----------
        model : str or dict
            The name, id, or dictionary representation of a model.

        Returns
        -------

        """
        rel = 'delete'

        filelist = cls.get_model_contents(model)
        for delfile in filelist:
            modelfileuri = cls.get_link(delfile, rel)
            delete(modelfileuri['uri'])

    @classmethod
    def copy_python_resources(cls, model):
        """Moves a model's score resources to the Compute server.

        Copies all of the Python score resources for a model to the pre-defined
        server location (/models/resources/viya/<model-UUID>/). To enable
        publishing and scoring, models that contain Python scoring resources
        need the score resource files to be copied to a set location
        (/models/resources/viya/<model-UUID>/). This location is used for
        integration with Event Stream Processing and others. This request
        invokes an asynchronous call to copy the score resource files. Check
        the individual score resource uris to get the completion state:
        pending, copying, success, failure. Please review the full Model
        Manager documentation before using.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model.

        Returns
        -------
        RestObj or None
            JSON response detailing the API metadata

        Warns
        -----
        UserWarning
            If no score resources exist for the model.

        """

        model_obj = cls.get_model(model)

        if model_obj is None:
            raise ValueError("No model '{}' found.".format(model))

        # Check if symbolic link for resource directories exists
        try:
            response = cls.put(
                '/models/%s/scoreResources' % model_obj['id'],
                headers={'Accept': 'application/json'},
            )
            if not response:
                warn("No score resources found for model '{}'".format(model_obj.name))
            return response
        except HTTPError as e:
            if e.code == 406:
                raise OSError(
                    'The SAS Viya system you are attempting to move score resources with requires an additional'
                    + ' administrator action in order to complete. Please see the documentation at '
                    + 'https://go.documentation.sas.com/doc/en/calcdc/3.5/calmodels/n10916nn7yro46n119nev9sb912c.htm,'
                    + 'which details the corollary approach for configuring analytic store model files.'
                )
            else:
                raise e

    @classmethod
    def convert_python_to_ds2(cls, model):
        """Converts a Python model to DS2

        For SAS Viya 3.5 Python models on SAS Model Manager, wrap the Python score code in DS2
        and convert the model score code type to DS2. Models converted in this way are not
        scoreable by CAS.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model.

        Returns
        -------
        API response
            JSON response detailing the API metadata

        """
        if cls.is_uuid(model):
            id_ = model
        elif isinstance(model, dict) and 'id' in model:
            id_ = model['id']
        else:
            model = cls.get_model(model)
            id_ = model['id']

        if isinstance(model, (str, dict)):
            model = cls.get_model(id_)

        ETag = model._headers['ETag']
        accept = 'text/vnd.sas.source.ds2'
        content = 'application/json'

        return cls.put(
            '/models/%s/typeConversion' % id_,
            headers={'Accept-Item': accept, 'Content-Type': content, 'If-Match': ETag},
        )

    @classmethod
    def get_model_details(cls, model):
        """Get model details from SAS Model Manager

        Get model details that pertain to model properties, model metadata,
        model input, output, and target variables, and user-defined values.

        Parameters
        ----------
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model.

        Returns
        -------
        API response
            JSON response detailing the model details

        """
        if cls.is_uuid(model):
            id_ = model
        elif isinstance(model, dict) and 'id' in model:
            id_ = model['id']
        else:
            model = cls.get_model(model)
            id_ = model['id']

        return cls.get('/models/%s' % id_)
